"""
PVs
# Last update Jan 2021 by RLO
# Now with Eiger goodness
"""

import os
import sys


def __show__(name):
    """Checks available variables given a string, uses first two letters"""
    for things in globals():
        if name[:2].lower() in things.lower():
            print("Available:", things)
    print()


def __which__():
    """Return script directory, used for finding which pv.py you are running"""
    pathname, scriptname = os.path.split(sys.argv[0])
    print("Current dir: " + os.path.abspath(pathname))
    print("path to pv.py: ")


requested_transmission = "BL24I-OP-ATTN-01:T2A:SETVAL1"

# PILATUS
pilat_filepath = "BL24I-EA-PILAT-01:cam1:FilePath"
pilat_filename = "BL24I-EA-PILAT-01:cam1:FileName"
pilat_filetemplate = "BL24I-EA-PILAT-01:cam1:FileTemplate"
pilat_numimages = "BL24I-EA-PILAT-01:cam1:NumImages"
pilat_numexpimage = "BL24I-EA-PILAT-01:cam1:NumExposures"
pilat_filenumber = "BL24I-EA-PILAT-01:cam1:FileNumber"
pilat_acquire = "BL24I-EA-PILAT-01:cam1:Acquire"
pilat_acquiretime = "BL24I-EA-PILAT-01:cam1:AcquireTime"
pilat_acquireperiod = "BL24I-EA-PILAT-01:cam1:AcquirePeriod"
pilat_imagemode = "BL24I-EA-PILAT-01:cam1:ImageMode"
pilat_triggermode = "BL24I-EA-PILAT-01:cam1:TriggerMode"
pilat_delaytime = "BL24I-EA-PILAT-01:cam1:DelayTime"
pilat_wavelength = "BL24I-EA-PILAT-01:cam1:Wavelength"
pilat_detdist = "BL24I-EA-PILAT-01:cam1:DetDist"
pilat_filtertrasm = "BL24I-EA-PILAT-01:cam1:FilterTransm"
pilat_filetemplate = "BL24I-EA-PILAT-01:cam1:FileTemplate"
pilat_beamx = "BL24I-EA-PILAT-01:cam1:BeamX"
pilat_beamy = "BL24I-EA-PILAT-01:cam1:BeamY"
pilat_startangle = "BL24I-EA-PILAT-01:cam1:StartAngle"
pilat_angleincr = "BL24I-EA-PILAT-01:cam1:AngleIncr"
pilat_omegaincr = "BL24I-EA-PILAT-01:cam1:OmegaIncr"
pilat_cbftemplate = "BL24I-EA-PILAT-01:cam1:CbfTemplateFile"
pilat_filenum = "BL24I-EA-PILAT-01:cam1:FileNumber_RBV"


# Eiger
eiger_filepath = "BL24I-EA-EIGER-01:CAM:FilePath"
eiger_filename = "BL24I-EA-EIGER-01:CAM:FileName"
eiger_ODfilepath = "BL24I-EA-EIGER-01:OD:FilePath"
eiger_ODfilename = "BL24I-EA-EIGER-01:OD:FileName"
eiger_seqID = "BL24I-EA-EIGER-01:CAM:SequenceId"
eiger_numimages = "BL24I-EA-EIGER-01:CAM:NumImages"
eiger_ODnumcapture = "BL24I-EA-EIGER-01:OD:NumCapture"
eiger_numexpimage = "BL24I-EA-EIGER-01:CAM:NumExposures"
eiger_acquiretime = "BL24I-EA-EIGER-01:CAM:AcquireTime"
eiger_acquireperiod = "BL24I-EA-EIGER-01:CAM:AcquirePeriod"
eiger_imagemode = "BL24I-EA-EIGER-01:CAM:ImageMode"
eiger_triggermode = "BL24I-EA-EIGER-01:CAM:TriggerMode"
eiger_numtriggers = "BL24I-EA-EIGER-01:CAM:NumTriggers"
eiger_manualtrigger = "BL24I-EA-EIGER-01:CAM:ManualTrigger"
eiger_trigger = "BL24I-EA-EIGER-01:CAM:Trigger"
eiger_filewriter = "BL24I-EA-EIGER-01:CAM:FWEnable"
eiger_stream = "BL24I-EA-EIGER-01:CAM:StreamEnable"
eiger_monitor = "BL24I-EA-EIGER-01:CAM:MonitorEnable"
eiger_datasource = "BL24I-EA-EIGER-01:CAM:DataSource"
eiger_statuspoll = "BL24I-EA-EIGER-01:CAM:ReadStatus.SCAN"
eiger_ROImode = "BL24I-EA-EIGER-01:CAM:ROIMode"
eiger_ff = "BL24I-EA-EIGER-01:CAM:FlatfieldApplied"
eiger_compress = "BL24I-EA-EIGER-01:CAM:FWCompression"
eiger_compresstype = "BL24I-EA-EIGER-01:CAM:CompressionAlgo"
eiger_ODcompress = "BL24I-EA-EIGER-01:OD:Compression"
eiger_ODdatatype = "BL24I-EA-EIGER-01:OD:DataType"
eiger_bitdepthrbv = "BL24I-EA-EIGER-01:CAM:BitDepthImage_RBV"
eiger_countmode = "BL24I-EA-EIGER-01:CAM:CountingMode"
eiger_autosum = "BL24I-EA-EIGER-01:CAM:AutoSummation"
eiger_hdrdetail = "BL24I-EA-EIGER-01:CAM:StreamHdrDetail"
eiger_hdrappen = "BL24I-EA-EIGER-01:CAM:StreamHdrAppendix"
eiger_ODcapture = "BL24I-EA-EIGER-01:OD:Capture"
eiger_acquire = "BL24I-EA-EIGER-01:CAM:Acquire"
eiger_wavelength = "BL24I-EA-EIGER-01:CAM:Wavelength"
eiger_detdist = "BL24I-EA-EIGER-01:CAM:DetDist"
eiger_beamx = "BL24I-EA-EIGER-01:CAM:BeamX"
eiger_beamy = "BL24I-EA-EIGER-01:CAM:BeamY"
eiger_omegaincr = "BL24I-EA-EIGER-01:CAM:OmegaIncr"
eiger_ODfilenameRBV = "BL24I-EA-EIGER-01:OD:FP:FileName_RBV"

# BPMs
qbpm1_inten = "BL24I-DI-QBPM-01:INTEN"
qbpm1_intenN = "BL24I-DI-QBPM-01:INTEN_N"
qbpm2_inten = "BL24I-DI-QBPM-02:INTEN"
qbpm2_intenN = "BL24I-DI-QBPM-02:INTEN_N"
qbpm3_inten = "BL24I-DI-QBPM-03:INTEN"
qbpm3_intenN = "BL24I-DI-QBPM-03:INTEN_N"
# Cividec
cividec_x = "BL24I-AL-XBPM-01:XS"
cividec_y = "BL24I-AL-XBPM-01:YS"
cividec_sumI = "BL24I-EA-XBPM-01:SumAll:Sigma_RBV"
cividec_beamx = "BL24I-EA-XBPM-01:PosX:MeanValue_RBV"
cividec_beamy = "BL24I-EA-XBPM-01:PosY:MeanValue_RBV"
# Lancelot
lance_x = "BL24I-EA-DET-03:X"
lance_y = "BL24I-EA-DET-03:Y"
lance_beamx = "BL24I-EA-DET-03:STAT:CentroidX_RBV"
lance_beamy = "BL24I-EA-DET-03:STAT:CentroidY_RBV"
# FAST SHUTTER
shtr_ctrl1 = "BL24I-EA-SHTR-01:CTRL1"
shtr_ctrl2 = "BL24I-EA-SHTR-01:CTRL2"
shtr_toggle = shtr_ctrl1
shtr_pos = shtr_ctrl2
# XPRESS3
xsp3_acquire = "BL24I-EA-XSP3-01:Acquire"
xsp3_erase = "BL24I-EA-XSP3-01:ERASE"
xsp3_acquiretime = "BL24I-EA-XSP3-01:AcquireTime"
xsp3_hdf5_filename = "BL24I-EA-XSP3-01:HDF5:FileName"
xsp3_hdf5_filepath = "BL24I-EA-XSP3-01:HDF5:FilePath"
xsp3_numimages = "BL24I-EA-XSP3-01:NumImages"
xsp3_triggermode = "BL24I-EA-XSP3-01:TriggerMode"
xsp3_c1_mca_roi1_llm = "BL24I-EA-XSP3-01:C1_MCA_ROI1_LLM"
xsp3_c1_mca_roi2_llm = "BL24I-EA-XSP3-01:C1_MCA_ROI2_LLM"
xsp3_c1_mca_roi3_llm = "BL24I-EA-XSP3-01:C1_MCA_ROI3_LLM"
xsp3_c1_mca_roi4_llm = "BL24I-EA-XSP3-01:C1_MCA_ROI4_LLM"
xsp3_c1_mca_roi1_hlm = "BL24I-EA-XSP3-01:C1_MCA_ROI1_HLM"
xsp3_c1_mca_roi2_hlm = "BL24I-EA-XSP3-01:C1_MCA_ROI2_HLM"
xsp3_c1_mca_roi3_hlm = "BL24I-EA-XSP3-01:C1_MCA_ROI3_HLM"
xsp3_c1_mca_roi4_hlm = "BL24I-EA-XSP3-01:C1_MCA_ROI4_HLM"
xsp3_c1_roi1_value_rbv = "BL24I-EA-XSP3-01:C1_ROI1:Value_RBV"
xsp3_c1_roi2_value_rbv = "BL24I-EA-XSP3-01:C1_ROI2:Value_RBV"
xsp3_c1_roi3_value_rbv = "BL24I-EA-XSP3-01:C1_ROI3:Value_RBV"
xsp3_c1_roi4_value_rbv = "BL24I-EA-XSP3-01:C1_ROI4:Value_RBV"
# Ring parameter
ring_current = "SR-DI-DCCT-01:SIGNAL"
ring_energy = "CS-CS-MSTAT-01:BEAMENERGY"
beam_energy = "BL24I-OP-DCM-01:ENERGY.RBV"
# Shutter status
port_shutter_status = "FE24I-PS-SHTR-01:STA"
optics_hutch_shutter_status = "FE24I-VA-GROUP-01:BLENABLE"
experimental_hutch_shutter_status = "BL24I-PS-SHTR-01:STA"
# S1
s1_x_gap = "BL24I-AL-SLITS-01:X:GAP"
s1_x_inboard = "BL24I-AL-SLITS-01:X:INBOARD"
s1_y_gap = "BL24I-AL-SLITS-01:Y:GAP"
s1_y_bottom = "BL24I-AL-SLITS-01:Y:BOTTOM"
# Mono
dcm_bragg = "BL24I-MO-DCM-01:BRAGG"
dcm_gap = "BL24I-MO-DCM-01:GAP"
dcm_roll1 = "BL24I-MO-DCM-01:XTAL1:ROLL"
dcm_roll2 = "BL24I-MO-DCM-01:XTAL2:ROLL"
dcm_pitch2 = "BL24I-MO-DCM-01:XTAL2:PITCH"
dcm_lambda = "BL24I-MO-DCM-01:LAMBDA"
dcm_energy = "BL24I-MO-DCM-01:ENERGY"

# S2
s2_x_plus = "BL24I-AL-SLITS-02:X:PLUS"
s2_x_minus = "BL24I-AL-SLITS-02:X:MINUS"
s2_y_plus = "BL24I-AL-SLITS-02:Y:PLUS"
s2_y_minus = "BL24I-AL-SLITS-02:Y:MINUS"
# PreFocussing Mirrors
hpfm_y1 = "BL24I-OP-HPFM-01:Y1"
hpfm_y2 = "BL24I-OP-HPFM-01:Y2"
hpfm_y3 = "BL24I-OP-HPFM-01:Y3"
hpfm_x1 = "BL24I-OP-HPFM-01:X1"
hpfm_x2 = "BL24I-OP-HPFM-01:X2"
vpfm_y1 = "BL24I-OP-VPFM-01:Y1"
vpfm_y2 = "BL24I-OP-VPFM-01:Y2"
vpfm_y3 = "BL24I-OP-VPFM-01:Y3"
vpfm_x1 = "BL24I-OP-VPFM-01:X1"
vpfm_x2 = "BL24I-OP-VPFM-01:X2"
# S3
s3_x_gap = "BL24I-AL-SLITS-03:X:GAP"
s3_x_inboard = "BL24I-AL-SLITS-03:X:INBOARD"
s3_y_gap = "BL24I-AL-SLITS-03:Y:GAP"
s3_y_top = "BL24I-AL-SLITS-03:Y:TOP"
# MicroFocussing Mirrors
hmfm_x = "BL24I-OP-HMFM-01:X"
hmfm_y = "BL24I-OP-HMFM-01:Y"
hmfm_pitch = "BL24I-OP-HMFM-01:PITCH"
vmfm_x = "BL24I-OP-VMFM-01:X"
vmfm_y = "BL24I-OP-VMFM-01:Y"
vmfm_pitch = "BL24I-OP-VMFM-01:PITCH"
mtab_z = "BL24I-OP-MTAB-01:Z"
# Collimation Table
ctab_x1 = "BL24I-MO-CTAB-01:X1"
ctab_x2 = "BL24I-MO-CTAB-01:X2"
# Attenuators
attn_disc1 = "BL24I-OP-ATTN-01:DISC1"
attn_disc2 = "BL24I-OP-ATTN-01:DISC2"
attn_match = "BL24I-OP-ATTN-01:MATCH"
# AP1
aptr1_x = "BL24I-AL-APTR-01:X"
aptr1_y = "BL24I-AL-APTR-01:Y"
aptr1_mp_select = "BL24I-AL-APTR-01:MP:SELECT"
# AP2
aptr2_x = "BL24I-AL-APTR-02:X"
aptr2_y = "BL24I-AL-APTR-02:Y"
aptr2_mp_select = "BL24I-AL-APTR-02:MP:SELECT"
# Vertical Pin Goniometer
vgon_omega = "BL24I-MO-VGON-01:OMEGA"
vgon_kappa = "BL24I-MO-VGON-01:KAPPA"
vgon_phi = "BL24I-MO-VGON-01:PHI"
vgon_pinxs = "BL24I-MO-VGON-01:PINXS"
vgon_pinys = "BL24I-MO-VGON-01:PINYS"
vgon_pinzs = "BL24I-MO-VGON-01:PINZS"
vgon_pinyh = "BL24I-MO-VGON-01:PINYH"
ptab_x = "BL24I-MO-PTAB-01:X"
ptab_z = "BL24I-MO-PTAB-01:Z"
ptab_y = "BL24I-MO-PTAB-01:Y"
# Horizontal Tray Goniometer
hgon_omega = "BL24I-MO-HGON-01:OMEGA"
hgon_trayys = "BL24I-MO-HGON-01:TRAYYS"
hgon_trayzs = "BL24I-MO-HGON-01:TRAYZS"
ttab_x = "BL24I-MO-TTAB-01:X"
ttab_y = "BL24I-MO-TTAB-01:Y"
ttab_z = "BL24I-MO-TTAB-01:Z"
# Cryo
cstrm_trans = "BL24I-MO-CSTRM-01:TRANS"
cstrm_mp_select = "BL24I-MO-CSTRM-01:MP:SELECT"
cstrm_p1701 = "BL24I-MO-CSTRM-01:P1701"
# Fluorescence Detector
fluo_trans = "BL24I-EA-DET-02:TRANS"
fluo_out_limit = "BL24I-EA-DET-02:OUT:LIMIT"
fluo_in_limit = "BL24I-EA-DET-02:IN:LIMIT"
# Beamstop
bs_x = "BL24I-MO-BS-01:X"
bs_y = "BL24I-MO-BS-01:Y"
bs_z = "BL24I-MO-BS-01:Z"
bs_roty = "BL24I-MO-BS-01:ROTY"
bs_mp_select = "BL24I-MO-BS-01:MP:SELECT"
# Backlight
bl_y = "BL24I-MO-BL-01:Y"
bl_mp_select = "BL24I-MO-BL-01:MP:SELECT"
# LED
led1_doxcurrent_ouputcurrent = "BL24I-DI-LED-01:DOXCURRENT:OUTPUTCURRENT"
led2_doxcurrent_ouputcurrent = "BL24I-DI-LED-02:DOXCURRENT:OUTPUTCURRENT"
led3_doxcurrent_ouputcurrent = "BL24I-DI-LED-03:DOXCURRENT:OUTPUTCURRENT"
led4_doxcurrent_ouputcurrent = "BL24I-DI-LED-04:DOXCURRENT:OUTPUTCURRENT"
# Detector
det_y = "BL24I-EA-DET-01:Y"
det_z = "BL24I-EA-DET-01:Z"
# Fast grid diagnostics
pmc_gridstatus = "BL24I-MO-STEP-10:signal:P2401"
pmc_gridcounter = "BL24I-MO-STEP-10:signal:P2402"
# PMAC Strings
step08_pmac_str = "BL24I-MO-IOC-08:ASYN8.AOUT"
step09_pmac_str = "BL24I-MO-IOC-09:ASYN9.AOUT"
step10_pmac_str = "BL24I-MO-IOC-10:ASYN10.AOUT"
step11_pmac_str = "BL24I-MO-IOC-11:ASYN11.AOUT"
step12_pmac_str = "BL24I-MO-IOC-12:ASYN12.AOUT"
step08_pmac_response = "BL24I-MO-IOC-08:ASYN8.AINP"
step09_pmac_response = "BL24I-MO-IOC-09:ASYN9.AINP"
step10_pmac_response = "BL24I-MO-IOC-10:ASYN10.AINP"
step11_pmac_response = "BL24I-MO-IOC-11:ASYN11.AINP"
step12_pmac_response = "BL24I-MO-IOC-12:ASYN12.AINP"
# General Purpose PV. Now use STEP13
ioc13_gp1 = "BL24I-MO-IOC-13:GP1"
ioc13_gp2 = "BL24I-MO-IOC-13:GP2"
ioc13_gp3 = "BL24I-MO-IOC-13:GP3"
ioc13_gp4 = "BL24I-MO-IOC-13:GP4"
ioc13_gp5 = "BL24I-MO-IOC-13:GP5"
ioc13_gp6 = "BL24I-MO-IOC-13:GP6"
ioc13_gp7 = "BL24I-MO-IOC-13:GP7"
ioc13_gp8 = "BL24I-MO-IOC-13:GP8"
ioc13_gp9 = "BL24I-MO-IOC-13:GP9"
ioc13_gp10 = "BL24I-MO-IOC-13:GP10"
ioc13_gp11 = "BL24I-MO-IOC-13:GP11"
ioc13_gp12 = "BL24I-MO-IOC-13:GP12"
ioc13_gp13 = "BL24I-MO-IOC-13:GP13"
ioc13_gp14 = "BL24I-MO-IOC-13:GP14"
ioc13_gp15 = "BL24I-MO-IOC-13:GP15"
# ME14E
me14e_pmac_str = "ME14E-MO-CHIP-01:PMAC_STRING"
me14e_stage_x = "ME14E-MO-CHIP-01:X"
me14e_stage_y = "ME14E-MO-CHIP-01:Y"
me14e_stage_z = "ME14E-MO-CHIP-01:Z"
me14e_filter = "ME14E-MO-CHIP-01:FILTER"
me14e_filepath = "ME14E-MO-CHIP-01:filePath"
me14e_chip_name = "ME14E-MO-CHIP-01:chipName"
me14e_chipcapacity = "ME14E-MO-CHIP-01:chipCapacity"
me14e_blockcapacity = "ME14E-MO-CHIP-01:blockCapacity"
me14e_exptime = "ME14E-MO-CHIP-01:expTime"
me14e_dcdetdist = "ME14E-MO-CHIP-01:detDistance"
me14e_scanstatus = "BL24I-MO-STEP-14:signal:P2401"
me14e_counter = "BL24I-MO-STEP-14:signal:P2402"
# ME14E General Purpose PV
me14e_gp1 = "ME14E-MO-IOC-01:GP1"
me14e_gp2 = "ME14E-MO-IOC-01:GP2"
me14e_gp3 = "ME14E-MO-IOC-01:GP3"
me14e_gp4 = "ME14E-MO-IOC-01:GP4"
me14e_gp5 = "ME14E-MO-IOC-01:GP5"
me14e_gp6 = "ME14E-MO-IOC-01:GP6"
me14e_gp7 = "ME14E-MO-IOC-01:GP7"
me14e_gp8 = "ME14E-MO-IOC-01:GP8"
me14e_gp9 = "ME14E-MO-IOC-01:GP9"
me14e_gp10 = "ME14E-MO-IOC-01:GP10"
me14e_gp11 = "ME14E-MO-IOC-01:GP11"
me14e_gp12 = "ME14E-MO-IOC-01:GP12"
me14e_gp13 = "ME14E-MO-IOC-01:GP13"
me14e_gp14 = "ME14E-MO-IOC-01:GP14"
me14e_gp15 = "ME14E-MO-IOC-01:GP15"
me14e_gp16 = "ME14E-MO-IOC-01:GP16"
me14e_gp17 = "ME14E-MO-IOC-01:GP17"
me14e_gp18 = "ME14E-MO-IOC-01:GP18"
me14e_gp19 = "ME14E-MO-IOC-01:GP19"
me14e_gp20 = "ME14E-MO-IOC-01:GP20"
me14e_gp21 = "ME14E-MO-IOC-01:GP21"
me14e_gp22 = "ME14E-MO-IOC-01:GP22"
me14e_gp23 = "ME14E-MO-IOC-01:GP23"
me14e_gp24 = "ME14E-MO-IOC-01:GP24"
me14e_gp25 = "ME14E-MO-IOC-01:GP25"
me14e_gp26 = "ME14E-MO-IOC-01:GP26"
me14e_gp27 = "ME14E-MO-IOC-01:GP27"
me14e_gp28 = "ME14E-MO-IOC-01:GP28"
me14e_gp29 = "ME14E-MO-IOC-01:GP29"
me14e_gp30 = "ME14E-MO-IOC-01:GP30"
me14e_gp31 = "ME14E-MO-IOC-01:GP31"
me14e_gp32 = "ME14E-MO-IOC-01:GP32"
me14e_gp33 = "ME14E-MO-IOC-01:GP33"
me14e_gp34 = "ME14E-MO-IOC-01:GP34"
me14e_gp35 = "ME14E-MO-IOC-01:GP35"
me14e_gp36 = "ME14E-MO-IOC-01:GP36"
me14e_gp37 = "ME14E-MO-IOC-01:GP37"
me14e_gp38 = "ME14E-MO-IOC-01:GP38"
me14e_gp39 = "ME14E-MO-IOC-01:GP39"
me14e_gp40 = "ME14E-MO-IOC-01:GP40"
me14e_gp41 = "ME14E-MO-IOC-01:GP41"
me14e_gp42 = "ME14E-MO-IOC-01:GP42"
me14e_gp43 = "ME14E-MO-IOC-01:GP43"
me14e_gp44 = "ME14E-MO-IOC-01:GP44"
me14e_gp45 = "ME14E-MO-IOC-01:GP45"
me14e_gp46 = "ME14E-MO-IOC-01:GP46"
me14e_gp47 = "ME14E-MO-IOC-01:GP47"
me14e_gp48 = "ME14E-MO-IOC-01:GP48"
me14e_gp49 = "ME14E-MO-IOC-01:GP49"
me14e_gp50 = "ME14E-MO-IOC-01:GP50"
me14e_gp51 = "ME14E-MO-IOC-01:GP51"
me14e_gp52 = "ME14E-MO-IOC-01:GP52"
me14e_gp53 = "ME14E-MO-IOC-01:GP53"
me14e_gp54 = "ME14E-MO-IOC-01:GP54"
me14e_gp55 = "ME14E-MO-IOC-01:GP55"
me14e_gp56 = "ME14E-MO-IOC-01:GP56"
me14e_gp57 = "ME14E-MO-IOC-01:GP57"
me14e_gp58 = "ME14E-MO-IOC-01:GP58"
me14e_gp59 = "ME14E-MO-IOC-01:GP59"
me14e_gp60 = "ME14E-MO-IOC-01:GP60"
me14e_gp61 = "ME14E-MO-IOC-01:GP61"
me14e_gp62 = "ME14E-MO-IOC-01:GP62"
me14e_gp63 = "ME14E-MO-IOC-01:GP63"
me14e_gp64 = "ME14E-MO-IOC-01:GP64"
me14e_gp65 = "ME14E-MO-IOC-01:GP65"
me14e_gp66 = "ME14E-MO-IOC-01:GP66"
me14e_gp67 = "ME14E-MO-IOC-01:GP67"
me14e_gp68 = "ME14E-MO-IOC-01:GP68"
me14e_gp69 = "ME14E-MO-IOC-01:GP69"
me14e_gp70 = "ME14E-MO-IOC-01:GP70"
me14e_gp71 = "ME14E-MO-IOC-01:GP71"
me14e_gp72 = "ME14E-MO-IOC-01:GP72"
me14e_gp73 = "ME14E-MO-IOC-01:GP73"
me14e_gp74 = "ME14E-MO-IOC-01:GP74"
me14e_gp75 = "ME14E-MO-IOC-01:GP75"
me14e_gp76 = "ME14E-MO-IOC-01:GP76"
me14e_gp77 = "ME14E-MO-IOC-01:GP77"
me14e_gp78 = "ME14E-MO-IOC-01:GP78"
me14e_gp79 = "ME14E-MO-IOC-01:GP79"
me14e_gp80 = "ME14E-MO-IOC-01:GP80"
me14e_gp81 = "ME14E-MO-IOC-01:GP81"
me14e_gp82 = "ME14E-MO-IOC-01:GP82"
me14e_gp83 = "ME14E-MO-IOC-01:GP83"
me14e_gp84 = "ME14E-MO-IOC-01:GP84"
me14e_gp85 = "ME14E-MO-IOC-01:GP85"
me14e_gp86 = "ME14E-MO-IOC-01:GP86"
me14e_gp87 = "ME14E-MO-IOC-01:GP87"
me14e_gp88 = "ME14E-MO-IOC-01:GP88"
me14e_gp89 = "ME14E-MO-IOC-01:GP89"
me14e_gp90 = "ME14E-MO-IOC-01:GP90"
me14e_gp91 = "ME14E-MO-IOC-01:GP91"
me14e_gp92 = "ME14E-MO-IOC-01:GP92"
me14e_gp93 = "ME14E-MO-IOC-01:GP93"
me14e_gp94 = "ME14E-MO-IOC-01:GP94"
me14e_gp95 = "ME14E-MO-IOC-01:GP95"
me14e_gp96 = "ME14E-MO-IOC-01:GP96"
me14e_gp97 = "ME14E-MO-IOC-01:GP97"
me14e_gp98 = "ME14E-MO-IOC-01:GP98"
me14e_gp99 = "ME14E-MO-IOC-01:GP99"
me14e_gp100 = "ME14E-MO-IOC-01:GP100"
me14e_gp101 = "ME14E-MO-IOC-01:GP101"  # Detector in use
me14e_gp102 = "ME14E-MO-IOC-01:GP102"
me14e_gp103 = "ME14E-MO-IOC-01:GP103"
me14e_gp104 = "ME14E-MO-IOC-01:GP104"
me14e_gp105 = "ME14E-MO-IOC-01:GP105"
me14e_gp106 = "ME14E-MO-IOC-01:GP106"
me14e_gp107 = "ME14E-MO-IOC-01:GP107"
me14e_gp108 = "ME14E-MO-IOC-01:GP108"
me14e_gp109 = "ME14E-MO-IOC-01:GP109"
me14e_gp110 = "ME14E-MO-IOC-01:GP110"
me14e_gp111 = "ME14E-MO-IOC-01:GP111"
me14e_gp112 = "ME14E-MO-IOC-01:GP112"
me14e_gp113 = "ME14E-MO-IOC-01:GP113"
me14e_gp114 = "ME14E-MO-IOC-01:GP114"
me14e_gp115 = "ME14E-MO-IOC-01:GP115"
me14e_gp116 = "ME14E-MO-IOC-01:GP116"
me14e_gp117 = "ME14E-MO-IOC-01:GP117"
me14e_gp118 = "ME14E-MO-IOC-01:GP118"
me14e_gp119 = "ME14E-MO-IOC-01:GP119"
me14e_gp120 = "ME14E-MO-IOC-01:GP120"



import argparse
from pathlib import Path
from unittest.mock import ANY, patch

from mx_bluesky.beamlines.i24.serial.run_serial import (
    get_location,
    run_extruder,
    run_fixed_target,
)


@patch("mx_bluesky.beamlines.i24.serial.run_serial.environ")
def test_get_location_not_on_beamline(mock_environ):
    mock_environ.get.return_value = None
    loc = get_location()

    assert loc == "dev"


@patch("mx_bluesky.beamlines.i24.serial.run_serial.subprocess")
@patch("mx_bluesky.beamlines.i24.serial.run_serial._get_file_path")
@patch("mx_bluesky.beamlines.i24.serial.run_serial.get_edm_path")
@patch("mx_bluesky.beamlines.i24.serial.run_serial._parse_input")
def test_run_fixed_extruder_in_test_mode(
    fake_parser, fake_edm_path, fake_filepath, mock_subprocess
):
    fake_parser.return_value = argparse.Namespace(test=True)
    fake_edm_path.return_value = Path("tmp/edm_serial")
    fake_filepath.return_value = Path("test/")

    run_extruder()

    mock_subprocess.run.assert_called_once_with(
        ["bash", Path("test/run_extruder.sh"), "tmp/edm_serial", "--test"]
    )


@patch("mx_bluesky.beamlines.i24.serial.run_serial.subprocess")
@patch("mx_bluesky.beamlines.i24.serial.run_serial.get_edm_path")
@patch("mx_bluesky.beamlines.i24.serial.run_serial._parse_input")
def test_run_fixed_target(fake_parser, fake_edm_path, mock_subprocess):
    fake_parser.return_value = argparse.Namespace(test=False)
    fake_edm_path.return_value = Path("tmp/edm_serial")

    run_fixed_target()

    mock_subprocess.run.assert_called_once_with(["bash", ANY, "tmp/edm_serial", ""])



from unittest.mock import MagicMock, patch

import pytest
from bluesky.preprocessors import run_decorator, subs_decorator
from ophyd_async.core import init_devices
from ophyd_async.epics.core import epics_signal_rw

from mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_callback import (
    GridscanISPyBCallback,
)
from mx_bluesky.common.parameters.constants import DocDescriptorNames
from mx_bluesky.common.plans.read_hardware import read_hardware_plan
from mx_bluesky.hyperion.parameters.gridscan import GridCommonWithHyperionDetectorParams

from .....conftest import (
    EXPECTED_START_TIME,
    TEST_DATA_COLLECTION_GROUP_ID,
    TEST_DATA_COLLECTION_IDS,
    TEST_SAMPLE_ID,
    TEST_SESSION_ID,
    TestData,
    assert_upsert_call_with,
    mx_acquisition_from_conn,
    remap_upsert_columns,
)

EXPECTED_DATA_COLLECTION_3D_XY = {
    "visitid": TEST_SESSION_ID,
    "parentid": TEST_DATA_COLLECTION_GROUP_ID,
    "sampleid": TEST_SAMPLE_ID,
    "comments": "MX-Bluesky: Xray centring 1 -",
    "detectorid": 78,
    "data_collection_number": 1,
    "detector_distance": 100.0,
    "exp_time": 0.1,
    "imgdir": "/tmp/",
    "imgprefix": "file_name",
    "imgsuffix": "h5",
    "n_passes": 1,
    "overlap": 0,
    "start_image_number": 1,
    "wavelength": None,
    "xbeam": 150.0,
    "ybeam": 160.0,
    "synchrotron_mode": None,
    "undulator_gap1": None,
    "starttime": EXPECTED_START_TIME,
    "filetemplate": "file_name_1_master.h5",
}

EXPECTED_DATA_COLLECTION_3D_XZ = EXPECTED_DATA_COLLECTION_3D_XY | {
    "comments": "MX-Bluesky: Xray centring 2 -",
    "data_collection_number": 2,
    "filetemplate": "file_name_2_master.h5",
}


TEST_GRID_INFO_IDS = (56, 57)
TEST_POSITION_ID = 78

EXPECTED_END_TIME = "2024-02-08 14:04:01"


@patch(
    "mx_bluesky.common.external_interaction.callbacks.common.ispyb_mapping.get_current_time_string",
    new=MagicMock(return_value=EXPECTED_START_TIME),
)
class TestXrayCentreISPyBCallback:
    def test_activity_gated_start_3d(self, mock_ispyb_conn):
        callback = GridscanISPyBCallback(
            param_type=GridCommonWithHyperionDetectorParams
        )
        callback.activity_gated_start(TestData.test_gridscan3d_start_document)  # pyright: ignore
        mx_acq = mx_acquisition_from_conn(mock_ispyb_conn)
        assert_upsert_call_with(
            mx_acq.upsert_data_collection_group.mock_calls[0],  # pyright: ignore
            mx_acq.get_data_collection_group_params(),
            {
                "parentid": TEST_SESSION_ID,
                "experimenttype": "Mesh3D",
                "sampleid": TEST_SAMPLE_ID,
            },
        )
        assert_upsert_call_with(
            mx_acq.upsert_data_collection.mock_calls[0],
            mx_acq.get_data_collection_params(),
            EXPECTED_DATA_COLLECTION_3D_XY,
        )
        assert_upsert_call_with(
            mx_acq.upsert_data_collection.mock_calls[1],
            mx_acq.get_data_collection_params(),
            EXPECTED_DATA_COLLECTION_3D_XZ,
        )
        mx_acq.upsert_data_collection.update_dc_position.assert_not_called()
        mx_acq.upsert_data_collection.upsert_dc_grid.assert_not_called()

    @patch(
        "mx_bluesky.common.external_interaction.ispyb.ispyb_store.StoreInIspyb.update_data_collection_group_table",
    )
    def test_reason_provided_if_crystal_not_found_error(
        self,
        mock_update_data_collection_group_table,
        mock_ispyb_conn,
    ):
        callback = GridscanISPyBCallback(
            param_type=GridCommonWithHyperionDetectorParams
        )
        callback.activity_gated_start(TestData.test_gridscan3d_start_document)  # pyright: ignore
        mx_acq = mx_acquisition_from_conn(mock_ispyb_conn)
        callback.activity_gated_stop(
            TestData.test_gridscan3d_stop_document_with_crystal_exception
        )
        assert mx_acq.update_data_collection_append_comments.call_args_list[0] == (
            (
                TEST_DATA_COLLECTION_IDS[0],
                "DataCollection Unsuccessful reason: Diffraction not found, skipping sample.",
                " ",
            ),
        )
        assert (
            mock_update_data_collection_group_table.call_args_list[0][0][0].comments
            == "Diffraction not found, skipping sample."
        )

    def test_hardware_read_event_3d(self, mock_ispyb_conn):
        callback = GridscanISPyBCallback(
            param_type=GridCommonWithHyperionDetectorParams
        )
        callback.activity_gated_start(TestData.test_gridscan3d_start_document)  # pyright: ignore
        mx_acq = mx_acquisition_from_conn(mock_ispyb_conn)
        mx_acq.upsert_data_collection_group.reset_mock()
        mx_acq.upsert_data_collection.reset_mock()
        callback.activity_gated_descriptor(
            TestData.test_descriptor_document_pre_data_collection
        )
        callback.activity_gated_event(TestData.test_event_document_pre_data_collection)
        mx_acq.upsert_data_collection_group.assert_called_once()
        expected_upsert = {
            "parentid": TEST_DATA_COLLECTION_GROUP_ID,
            "slitgaphorizontal": 0.1234,
            "slitgapvertical": 0.2345,
            "synchrotronmode": "User",
            "undulatorgap1": 1.234,
            "resolution": 1.1830593331191241,
            "wavelength": 1.11647184541378,
        }
        assert_upsert_call_with(
            mx_acq.upsert_data_collection.mock_calls[0],
            mx_acq.get_data_collection_params(),
            {"id": TEST_DATA_COLLECTION_IDS[0], **expected_upsert},
        )
        assert_upsert_call_with(
            mx_acq.upsert_data_collection.mock_calls[1],
            mx_acq.get_data_collection_params(),
            {"id": TEST_DATA_COLLECTION_IDS[1], **expected_upsert},
        )

    def test_flux_read_events_3d(self, mock_ispyb_conn):
        callback = GridscanISPyBCallback(
            param_type=GridCommonWithHyperionDetectorParams
        )
        callback.activity_gated_start(TestData.test_gridscan3d_start_document)  # pyright: ignore
        mx_acq = mx_acquisition_from_conn(mock_ispyb_conn)
        callback.activity_gated_descriptor(
            TestData.test_descriptor_document_pre_data_collection
        )
        callback.activity_gated_event(TestData.test_event_document_pre_data_collection)
        mx_acq.upsert_data_collection_group.reset_mock()
        mx_acq.upsert_data_collection.reset_mock()

        callback.activity_gated_descriptor(
            TestData.test_descriptor_document_during_data_collection
        )
        callback.activity_gated_event(
            TestData.test_event_document_during_data_collection
        )

        assert_upsert_call_with(
            mx_acq.upsert_data_collection.mock_calls[0],
            mx_acq.get_data_collection_params(),
            {
                "parentid": TEST_DATA_COLLECTION_GROUP_ID,
                "id": TEST_DATA_COLLECTION_IDS[0],
                "wavelength": 1.11647184541378,
                "transmission": 100,
                "flux": 10,
                "resolution": 1.1830593331191241,
                "focal_spot_size_at_samplex": 0.05,
                "focal_spot_size_at_sampley": 0.02,
                "beamsize_at_samplex": 0.05,
                "beamsize_at_sampley": 0.02,
            },
        )
        assert_upsert_call_with(
            mx_acq.upsert_data_collection.mock_calls[1],
            mx_acq.get_data_collection_params(),
            {
                "parentid": TEST_DATA_COLLECTION_GROUP_ID,
                "id": TEST_DATA_COLLECTION_IDS[1],
                "wavelength": 1.11647184541378,
                "transmission": 100,
                "flux": 10,
                "resolution": 1.1830593331191241,
                "focal_spot_size_at_samplex": 0.05,
                "focal_spot_size_at_sampley": 0.02,
                "beamsize_at_samplex": 0.05,
                "beamsize_at_sampley": 0.02,
            },
        )
        mx_acq.update_dc_position.assert_not_called()
        mx_acq.upsert_dc_grid.assert_not_called()

    def test_activity_gated_event_oav_snapshot_triggered(self, mock_ispyb_conn):
        callback = GridscanISPyBCallback(
            param_type=GridCommonWithHyperionDetectorParams
        )
        callback.activity_gated_start(TestData.test_gridscan3d_start_document)  # pyright: ignore
        mx_acq = mx_acquisition_from_conn(mock_ispyb_conn)
        mx_acq.upsert_data_collection_group.reset_mock()
        mx_acq.upsert_data_collection.reset_mock()

        callback.activity_gated_descriptor(
            TestData.test_descriptor_document_oav_snapshot
        )
        callback.activity_gated_event(TestData.test_event_document_oav_snapshot_xy)
        callback.activity_gated_event(TestData.test_event_document_oav_snapshot_xz)

        assert_upsert_call_with(
            mx_acq.upsert_data_collection.mock_calls[0],
            mx_acq.get_data_collection_params(),
            {
                "id": TEST_DATA_COLLECTION_IDS[0],
                "parentid": TEST_DATA_COLLECTION_GROUP_ID,
                "nimages": 40 * 20,
                "xtal_snapshot1": "test_1_y",
                "xtal_snapshot2": "test_2_y",
                "xtal_snapshot3": "test_3_y",
                "axisstart": 0,
                "omegastart": 0,
                "axisend": 0,
                "axisrange": 0,
            },
        )
        mx_acq.update_data_collection_append_comments.assert_any_call(
            TEST_DATA_COLLECTION_IDS[0],
            "Diffraction grid scan of 40 by 20 "
            "images in 126.4 um by 126.4 um steps. Top left (px): [50,100], "
            "bottom right (px): [3250,1700].",
            " ",
        )
        assert_upsert_call_with(
            mx_acq.upsert_data_collection.mock_calls[1],
            mx_acq.get_data_collection_params(),
            {
                "id": TEST_DATA_COLLECTION_IDS[1],
                "parentid": TEST_DATA_COLLECTION_GROUP_ID,
                "nimages": 40 * 10,
                "xtal_snapshot1": "test_1_z",
                "xtal_snapshot2": "test_2_z",
                "xtal_snapshot3": "test_3_z",
                "axisstart": 90,
                "omegastart": 90,
                "axisend": 90,
                "axisrange": 0,
            },
        )
        mx_acq.update_data_collection_append_comments.assert_any_call(
            TEST_DATA_COLLECTION_IDS[1],
            "Diffraction grid scan of 40 by 10 "
            "images in 126.4 um by 126.4 um steps. Top left (px): [50,0], "
            "bottom right (px): [3250,800].",
            " ",
        )
        assert_upsert_call_with(
            mx_acq.upsert_dc_grid.mock_calls[0],
            mx_acq.get_dc_grid_params(),
            {
                "parentid": TEST_DATA_COLLECTION_IDS[0],
                "dxinmm": 0.1264,
                "dyinmm": 0.1264,
                "stepsx": 40,
                "stepsy": 20,
                "micronsperpixelx": 1.58,
                "micronsperpixely": 1.58,
                "snapshotoffsetxpixel": 50,
                "snapshotoffsetypixel": 100,
                "orientation": "horizontal",
                "snaked": True,
            },
        )
        assert_upsert_call_with(
            mx_acq.upsert_dc_grid.mock_calls[1],
            mx_acq.get_dc_grid_params(),
            {
                "parentid": TEST_DATA_COLLECTION_IDS[1],
                "dxinmm": 0.1264,
                "dyinmm": 0.1264,
                "stepsx": 40,
                "stepsy": 10,
                "micronsperpixelx": 1.58,
                "micronsperpixely": 1.58,
                "snapshotoffsetxpixel": 50,
                "snapshotoffsetypixel": 0,
                "orientation": "horizontal",
                "snaked": True,
            },
        )

        group_dc_cols = remap_upsert_columns(
            mx_acq.get_data_collection_group_params(),
            mx_acq.upsert_data_collection_group.mock_calls[1].args[0],
        )
        assert group_dc_cols["comments"] == "Diffraction grid scan of 40 by 20 by 10."

    async def test_ispyb_callback_handles_read_hardware_in_run_engine(
        self, RE, mock_ispyb_conn, dummy_rotation_data_collection_group_info
    ):
        callback = GridscanISPyBCallback(
            param_type=GridCommonWithHyperionDetectorParams
        )
        callback._handle_ispyb_hardware_read = MagicMock()
        callback._handle_ispyb_transmission_flux_read = MagicMock()
        callback.ispyb = MagicMock()
        callback.params = MagicMock()
        callback.data_collection_group_info = dummy_rotation_data_collection_group_info

        with init_devices(mock=True):
            test_readable = epics_signal_rw(str, "pv")

        @subs_decorator(callback)
        @run_decorator(
            md={
                "activate_callbacks": ["GridscanISPyBCallback"],
            },
        )
        def test_plan():
            yield from read_hardware_plan(
                [test_readable], DocDescriptorNames.HARDWARE_READ_PRE
            )
            yield from read_hardware_plan(
                [test_readable], DocDescriptorNames.HARDWARE_READ_DURING
            )

        RE(test_plan())

        callback._handle_ispyb_hardware_read.assert_called_once()
        callback._handle_ispyb_transmission_flux_read.assert_called_once()

    @patch(
        "mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_callback.GridscanISPyBCallback._handle_zocalo_read_event",
    )
    @patch(
        "mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_callback.GridscanISPyBCallback._handle_oav_grid_snapshot_triggered",
    )
    @patch(
        "mx_bluesky.common.external_interaction.ispyb.ispyb_store.StoreInIspyb.update_deposition",
    )
    @patch(
        "mx_bluesky.common.external_interaction.ispyb.ispyb_store.StoreInIspyb.update_data_collection_group_table",
    )
    def test_given_event_doc_before_start_doc_recieved_then_exception_raised(
        self,
        mock_update_data_collection_group_table,
        mock_update_deposition,
        mock__handle_oav_grid_snapshot_triggered,
        mock__handle_zocalo_read_event,
    ):
        callback = GridscanISPyBCallback(
            param_type=GridCommonWithHyperionDetectorParams
        )
        callback.activity_gated_descriptor(
            TestData.test_descriptor_document_oav_snapshot
        )
        callback.ispyb = MagicMock()
        callback.params = MagicMock()
        callback.data_collection_group_info = None
        with pytest.raises(AssertionError) as e:
            callback.activity_gated_event(TestData.test_event_document_oav_snapshot_xy)

        assert "No data collection group info" in str(e.value)



from __future__ import annotations

from mx_bluesky.common.external_interaction.ispyb.data_model import DataCollectionInfo
from mx_bluesky.hyperion.parameters.rotation import RotationScan


def populate_data_collection_info_for_rotation(params: RotationScan):
    info = DataCollectionInfo(
        omega_start=params.omega_start_deg,
        data_collection_number=params.detector_params.run_number,  # type:ignore # the validator always makes this int
        n_images=params.num_images,
        axis_range=params.rotation_increment_deg,
        axis_start=params.omega_start_deg,
        axis_end=(params.omega_start_deg + params.scan_width_deg),
        kappa_start=params.kappa_start_deg,
    )
    return info



import numpy
import pytest
from bluesky.simulators import RunEngineSimulator, assert_message_and_return_remaining
from dodal.devices.aperturescatterguard import ApertureScatterguard, ApertureValue
from dodal.devices.smargon import Smargon, StubPosition

from mx_bluesky.common.xrc_result import XRayCentreResult
from mx_bluesky.hyperion.experiment_plans.change_aperture_then_move_plan import (
    change_aperture_then_move_to_xtal,
)
from mx_bluesky.hyperion.parameters.gridscan import HyperionSpecifiedThreeDGridScan


@pytest.fixture
def simple_flyscan_hit():
    return XRayCentreResult(
        centre_of_mass_mm=numpy.array([0.1, 0.2, 0.3]),
        bounding_box_mm=(
            numpy.array([0.09, 0.19, 0.29]),
            numpy.array([0.11, 0.21, 0.31]),
        ),
        max_count=20,
        total_count=57,
    )


@pytest.mark.parametrize("set_stub_offsets", [True, False])
def test_change_aperture_then_move_to_xtal_happy_path(
    sim_run_engine: RunEngineSimulator,
    simple_flyscan_hit: XRayCentreResult,
    smargon: Smargon,
    aperture_scatterguard: ApertureScatterguard,
    test_fgs_params: HyperionSpecifiedThreeDGridScan,
    set_stub_offsets: bool,
):
    test_fgs_params.features.set_stub_offsets = set_stub_offsets
    msgs = sim_run_engine.simulate_plan(
        change_aperture_then_move_to_xtal(
            simple_flyscan_hit, smargon, aperture_scatterguard, test_fgs_params
        )
    )

    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj is aperture_scatterguard
        and msg.args[0] == ApertureValue.MEDIUM,
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj is smargon.x
        and msg.args[0] == 0.1,
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj is smargon.y
        and msg.args[0] == 0.2,
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj is smargon.z
        and msg.args[0] == 0.3,
    )
    if set_stub_offsets:
        assert_message_and_return_remaining(
            msgs,
            lambda msg: msg.command == "set"
            and msg.obj is smargon.stub_offsets
            and msg.args[0] == StubPosition.CURRENT_AS_CENTER,
        )
    else:
        assert all(
            not (msg.command == "set" and msg.obj is smargon.stub_offsets)
            for msg in msgs
        )



from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import cast

import bluesky.plan_stubs as bps
from bluesky.utils import MsgGenerator
from dodal.common.beamlines.beamline_utils import get_path_provider
from dodal.common.types import UpdatingPathProvider
from dodal.devices.fast_grid_scan import PandAGridScanParams
from dodal.devices.smargon import Smargon
from ophyd_async.fastcs.panda import (
    HDFPanda,
    SeqTable,
    SeqTrigger,
)

from mx_bluesky.common.device_setup_plans.setup_panda import load_panda_from_yaml
from mx_bluesky.common.utils.log import LOGGER
from mx_bluesky.hyperion.parameters.constants import DeviceSettingsConstants

MM_TO_ENCODER_COUNTS = 200000
GENERAL_TIMEOUT = 60
TICKS_PER_MS = 1000  # Panda sequencer prescaler will be set to us
PULSE_WIDTH_US = 50


class Enabled(Enum):
    ENABLED = "ONE"
    DISABLED = "ZERO"


class PcapArm(Enum):
    ARMED = "Arm"
    DISARMED = "Disarm"


def _get_seq_table(
    parameters: PandAGridScanParams, exposure_distance_mm, time_between_steps_ms
) -> SeqTable:
    """
    Generate the sequencer table for the panda.

    - Sending a 'trigger' means trigger PCAP internally and send signal to Eiger via physical panda output

    SEQUENCER TABLE:

        1. Wait for physical trigger from motion script to mark start of scan / change of direction
        2. Wait for POSA (X2) to be greater than X_START and send x_steps triggers every time_between_steps_ms
        3. Wait for physical trigger from motion script to mark change of direction
        4. Wait for POSA (X2) to be less than X_START + X_STEP_SIZE * x_steps + exposure distance, then
            send x_steps triggers every time_between_steps_ms
        5. Go back to step one.

        For a more detailed explanation and a diagram, see https://github.com/DiamondLightSource/hyperion/wiki/PandA-constant%E2%80%90motion-scanning

        For documentation on Panda itself, see https://pandablocks.github.io/PandABlocks-FPGA/master/index.html

    Args:
        exposure_distance_mm: The distance travelled by the sample each time the detector is exposed: exposure time * sample velocity
        time_between_steps_ms: The time taken to traverse between each grid step.
        parameters: Parameters for the panda gridscan

    Returns:
        An instance of SeqTable describing the panda sequencer table
    """

    start_of_grid_x_counts = int(parameters.x_start_mm * MM_TO_ENCODER_COUNTS)

    # x_start is the first trigger point, so we need to travel to x_steps-1 for the final trigger point
    end_of_grid_x_counts = int(
        start_of_grid_x_counts
        + (parameters.x_step_size_mm * (parameters.x_steps - 1) * MM_TO_ENCODER_COUNTS)
    )

    exposure_distance_x_counts = int(exposure_distance_mm * MM_TO_ENCODER_COUNTS)

    num_pulses = parameters.x_steps

    # Integer precision here is 1e-6s, so casting is safe
    delay_between_pulses = int(time_between_steps_ms * TICKS_PER_MS)

    assert delay_between_pulses > PULSE_WIDTH_US

    # BITA_1 trigger wired from TTLIN1, this is the trigger input

    # +ve direction scan

    table = (
        SeqTable.row(trigger=SeqTrigger.BITA_1, time2=1)
        + SeqTable.row(
            repeats=num_pulses,
            trigger=SeqTrigger.POSA_GT,
            position=start_of_grid_x_counts,
            time1=PULSE_WIDTH_US,
            outa1=True,
            time2=delay_between_pulses - PULSE_WIDTH_US,
            outa2=False,
        )
        +
        # -ve direction scan
        SeqTable.row(trigger=SeqTrigger.BITA_1, time2=1)
        + SeqTable.row(
            repeats=num_pulses,
            trigger=SeqTrigger.POSA_LT,
            position=end_of_grid_x_counts + exposure_distance_x_counts,
            time1=PULSE_WIDTH_US,
            outa1=True,
            time2=delay_between_pulses - PULSE_WIDTH_US,
            outa2=False,
        )
    )

    return table


def setup_panda_for_flyscan(
    panda: HDFPanda,
    parameters: PandAGridScanParams,
    smargon: Smargon,
    exposure_time_s: float,
    time_between_x_steps_ms: float,
    sample_velocity_mm_per_s: float,
) -> MsgGenerator:
    """Configures the PandA device for a flyscan.
    Sets PVs from a yaml file, calibrates the encoder, and
    adjusts the sequencer table based off the grid parameters. Yaml file can be
    created using ophyd_async.core.save_device()

    Args:
        panda (HDFPanda): The PandA Ophyd device
        parameters (PandAGridScanParams): Grid parameters
        smargon (Smargon): The Smargon Ophyd device
        exposure_time_s (float): Detector exposure time per trigger
        time_between_x_steps_ms (float): Time, in ms, between each trigger. Equal to deadtime + exposure time
        sample_velocity_mm_per_s (float): Velocity of the sample in mm/s = x_step_size_mm * 1000 /
            time_between_x_steps_ms
    Returns:
        MsgGenerator

    Yields:
        Iterator[MsgGenerator]
    """
    assert parameters.x_steps > 0
    assert time_between_x_steps_ms * 1000 >= exposure_time_s
    assert sample_velocity_mm_per_s * exposure_time_s < parameters.x_step_size_mm

    yield from bps.stage(panda, group="panda-config")

    yield from load_panda_from_yaml(
        DeviceSettingsConstants.PANDA_FLYSCAN_SETTINGS_DIR,
        DeviceSettingsConstants.PANDA_FLYSCAN_SETTINGS_FILENAME,
        panda,
    )

    initial_x = yield from bps.rd(smargon.x.user_readback)
    initial_y = yield from bps.rd(smargon.y.user_readback)
    initial_z = yield from bps.rd(smargon.z.user_readback)

    # Home the PandA X, Y, and Z encoders using current motor position
    yield from bps.abs_set(
        panda.inenc[1].setp,  # type: ignore
        initial_x * MM_TO_ENCODER_COUNTS,
        wait=True,
    )

    yield from bps.abs_set(
        panda.inenc[2].setp,  # type: ignore
        initial_y * MM_TO_ENCODER_COUNTS,
        wait=True,
    )

    yield from bps.abs_set(
        panda.inenc[3].setp,  # type: ignore
        initial_z * MM_TO_ENCODER_COUNTS,
        wait=True,
    )

    yield from bps.abs_set(panda.pulse[1].width, exposure_time_s, group="panda-config")

    exposure_distance_mm = sample_velocity_mm_per_s * exposure_time_s

    table = _get_seq_table(parameters, exposure_distance_mm, time_between_x_steps_ms)

    yield from bps.abs_set(panda.seq[1].table, table, group="panda-config")

    yield from bps.abs_set(
        panda.pcap.enable,  # type: ignore
        Enabled.ENABLED.value,
        group="panda-config",
    )

    # Values need to be set before blocks are enabled, so wait here
    yield from bps.wait(group="panda-config", timeout=GENERAL_TIMEOUT)

    LOGGER.info(f"PandA sequencer table has been set to: {str(table)}")
    table_readback = yield from bps.rd(panda.seq[1].table)
    LOGGER.debug(f"PandA sequencer table readback is: {str(table_readback)}")

    yield from arm_panda_for_gridscan(panda)


def arm_panda_for_gridscan(panda: HDFPanda, group="arm_panda_gridscan"):
    yield from bps.abs_set(panda.seq[1].enable, Enabled.ENABLED.value, group=group)  # type: ignore
    yield from bps.abs_set(panda.pulse[1].enable, Enabled.ENABLED.value, group=group)  # type: ignore
    yield from bps.abs_set(panda.counter[1].enable, Enabled.ENABLED.value, group=group)  # type: ignore
    yield from bps.abs_set(panda.pcap.arm, PcapArm.ARMED.value, group=group)  # type: ignore
    yield from bps.wait(group=group, timeout=GENERAL_TIMEOUT)
    LOGGER.info("PandA has been armed")


def disarm_panda_for_gridscan(panda, group="disarm_panda_gridscan") -> MsgGenerator:
    yield from bps.abs_set(panda.pcap.arm, PcapArm.DISARMED.value, group=group)  # type: ignore
    yield from bps.abs_set(panda.counter[1].enable, Enabled.DISABLED.value, group=group)  # type: ignore
    yield from bps.abs_set(panda.seq[1].enable, Enabled.DISABLED.value, group=group)
    yield from bps.abs_set(panda.pulse[1].enable, Enabled.DISABLED.value, group=group)
    yield from bps.abs_set(panda.pcap.enable, Enabled.DISABLED.value, group=group)  # type: ignore
    yield from bps.wait(group=group, timeout=GENERAL_TIMEOUT)


def set_panda_directory(panda_directory: Path) -> MsgGenerator:
    """Updates the root folder which is used by the PandA's PCAP."""

    suffix = datetime.now().strftime("_%Y%m%d%H%M%S")

    async def set_panda_dir():
        await cast(UpdatingPathProvider, get_path_provider()).update(
            directory=panda_directory, suffix=suffix
        )

    yield from bps.wait_for([set_panda_dir])



from copy import deepcopy
from unittest.mock import MagicMock, patch

import numpy as np
import pytest
from numpy.typing import DTypeLike

from mx_bluesky.common.external_interaction.callbacks.xray_centre.nexus_callback import (
    GridscanNexusFileCallback,
)
from mx_bluesky.hyperion.parameters.gridscan import HyperionSpecifiedThreeDGridScan

from .....conftest import TestData


@pytest.fixture
def nexus_writer():
    with patch(
        "mx_bluesky.common.external_interaction.nexus.write_nexus.NexusWriter"
    ) as nw:
        yield nw


def test_writers_not_sDTypeLikeetup_on_plan_start_doc(
    nexus_writer: MagicMock,
):
    nexus_handler = GridscanNexusFileCallback(
        param_type=HyperionSpecifiedThreeDGridScan
    )
    nexus_writer.assert_not_called()
    nexus_handler.activity_gated_start(TestData.test_start_document)
    nexus_writer.assert_not_called()


@patch(
    "mx_bluesky.common.external_interaction.callbacks.xray_centre.nexus_callback.NexusWriter"
)
def test_writers_dont_create_on_init_but_do_on_during_collection_read_event(
    mock_nexus_writer: MagicMock,
):
    mock_nexus_writer.side_effect = [MagicMock(), MagicMock()]
    nexus_handler = GridscanNexusFileCallback(
        param_type=HyperionSpecifiedThreeDGridScan
    )

    assert nexus_handler.nexus_writer_1 is None
    assert nexus_handler.nexus_writer_2 is None

    nexus_handler.activity_gated_start(TestData.test_gridscan_outer_start_document)  # type: ignore
    nexus_handler.activity_gated_descriptor(
        TestData.test_descriptor_document_during_data_collection
    )

    nexus_handler.activity_gated_event(
        TestData.test_event_document_during_data_collection
    )

    assert nexus_handler.nexus_writer_1 is not None
    assert nexus_handler.nexus_writer_2 is not None
    nexus_handler.nexus_writer_1.create_nexus_file.assert_called_once()
    nexus_handler.nexus_writer_2.create_nexus_file.assert_called_once()


@pytest.mark.parametrize(
    ["bit_depth", "vds_type"],
    [
        (8, np.uint8),
        (16, np.uint16),
        (32, np.uint32),
    ],
)
@patch(
    "mx_bluesky.common.external_interaction.callbacks.xray_centre.nexus_callback.NexusWriter"
)
def test_given_different_bit_depths_then_writers_created_wth_correct_VDS_size(
    mock_nexus_writer: MagicMock,
    bit_depth: int,
    vds_type: DTypeLike,
):
    mock_nexus_writer.side_effect = [MagicMock(), MagicMock()]
    nexus_handler = GridscanNexusFileCallback(
        param_type=HyperionSpecifiedThreeDGridScan
    )

    nexus_handler.activity_gated_start(TestData.test_start_document)
    nexus_handler.activity_gated_descriptor(
        TestData.test_descriptor_document_during_data_collection
    )
    event_doc = deepcopy(TestData.test_event_document_during_data_collection)
    event_doc["data"]["eiger_bit_depth"] = bit_depth

    nexus_handler.activity_gated_event(event_doc)

    assert nexus_handler.nexus_writer_1 is not None
    assert nexus_handler.nexus_writer_2 is not None
    nexus_handler.nexus_writer_1.create_nexus_file.assert_called_once_with(  # type:ignore
        vds_type
    )
    nexus_handler.nexus_writer_2.create_nexus_file.assert_called_once_with(  # type:ignore
        vds_type
    )


@patch(
    "mx_bluesky.common.external_interaction.callbacks.xray_centre.nexus_callback.NexusWriter"
)
def test_beam_and_attenuator_set_on_ispyb_transmission_event(
    mock_nexus_writer: MagicMock,
):
    mock_nexus_writer.side_effect = [MagicMock(), MagicMock()]
    nexus_handler = GridscanNexusFileCallback(
        param_type=HyperionSpecifiedThreeDGridScan
    )

    nexus_handler.activity_gated_start(TestData.test_start_document)
    nexus_handler.activity_gated_descriptor(
        TestData.test_descriptor_document_during_data_collection
    )
    nexus_handler.activity_gated_event(
        TestData.test_event_document_during_data_collection
    )

    for writer in [nexus_handler.nexus_writer_1, nexus_handler.nexus_writer_2]:
        assert writer is not None
        assert writer.attenuator is not None
        assert writer.beam is not None


def test_sensible_error_if_writing_triggered_before_params_received(
    nexus_writer: MagicMock,
):
    nexus_handler = GridscanNexusFileCallback(
        param_type=HyperionSpecifiedThreeDGridScan
    )
    nexus_handler.activity_gated_descriptor(
        TestData.test_descriptor_document_during_data_collection
    )
    with pytest.raises(AssertionError) as excinfo:
        nexus_handler.activity_gated_event(
            TestData.test_event_document_during_data_collection
        )

    assert "Nexus callback did not receive start doc" in excinfo.value.args[0]



from collections.abc import AsyncGenerator
from functools import partial
from unittest.mock import ANY, AsyncMock, MagicMock, call, patch

import pytest
from bluesky.run_engine import RunEngine
from bluesky.simulators import assert_message_and_return_remaining
from dodal.beamlines import i04
from dodal.devices.oav.oav_detector import OAV, OAVConfig
from dodal.devices.oav.oav_to_redis_forwarder import OAVToRedisForwarder, Source
from dodal.devices.robot import BartRobot
from dodal.devices.smargon import Smargon
from dodal.devices.thawer import Thawer, ThawerStates
from ophyd.sim import NullStatus
from ophyd_async.core import (
    AsyncStatus,
    init_devices,
)
from ophyd_async.epics.motor import Motor
from ophyd_async.testing import (
    callback_on_mock_put,
    get_mock_put,
    set_mock_value,
)

from mx_bluesky.beamlines.i04.thawing_plan import (
    thaw,
    thaw_and_stream_to_redis,
)

DISPLAY_CONFIGURATION = "tests/test_data/test_display.configuration"
ZOOM_LEVELS_XML = "tests/test_data/test_jCameraManZoomLevels.xml"


class MyException(Exception):
    pass


def patch_motor(motor: Motor, initial_position: float = 0):
    set_mock_value(motor.user_setpoint, initial_position)
    set_mock_value(motor.user_readback, initial_position)
    set_mock_value(motor.deadband, 0.001)
    set_mock_value(motor.motor_done_move, 1)
    set_mock_value(motor.velocity, 3)
    return callback_on_mock_put(
        motor.user_setpoint,
        lambda pos, *args, **kwargs: set_mock_value(motor.user_readback, pos),
    )


@pytest.fixture
async def oav(RE: RunEngine) -> OAV:
    oav_config = OAVConfig(ZOOM_LEVELS_XML, DISPLAY_CONFIGURATION)
    async with init_devices(mock=True, connect=True):
        oav = OAV("", config=oav_config, name="fake_oav")
    zoom_levels_list = ["1.0x", "2.0x", "5.0x"]
    oav.zoom_controller._get_allowed_zoom_levels = AsyncMock(
        return_value=zoom_levels_list
    )
    set_mock_value(oav.zoom_controller.level, "1.0x")
    set_mock_value(oav.grid_snapshot.x_size, 1024)
    set_mock_value(oav.grid_snapshot.y_size, 768)
    return oav


@pytest.fixture
async def smargon(RE: RunEngine) -> AsyncGenerator[Smargon, None]:
    smargon = Smargon(name="smargon")
    await smargon.connect(mock=True)

    set_mock_value(smargon.omega.user_readback, 0.0)

    with patch_motor(smargon.omega):
        yield smargon


@pytest.fixture
def thawer(RE: RunEngine) -> Thawer:
    return i04.thawer(connect_immediately=True, mock=True)


@pytest.fixture
async def oav_forwarder(RE: RunEngine) -> OAVToRedisForwarder:
    with init_devices(mock=True):
        oav_forwarder = OAVToRedisForwarder(
            "prefix", "host", "password", name="oav_to_redis_forwarder"
        )

    # Replace when https://github.com/bluesky/ophyd-async/issues/521 is released
    @AsyncStatus.wrap
    async def completed_status():
        pass

    oav_forwarder.kickoff = MagicMock(side_effect=completed_status)
    oav_forwarder.complete = MagicMock(side_effect=completed_status)
    return oav_forwarder


@pytest.fixture
def robot(RE: RunEngine) -> BartRobot:
    return i04.robot(connect_immediately=True, mock=True)


def _do_thaw_and_confirm_cleanup(
    move_mock: MagicMock, smargon: Smargon, thawer: Thawer, do_thaw_func
):
    smargon.omega.set = move_mock
    set_mock_value(smargon.omega.velocity, initial_velocity := 10)
    smargon.omega.set = move_mock
    do_thaw_func()
    last_thawer_call = get_mock_put(thawer.control).call_args_list[-1]
    assert last_thawer_call == call(ThawerStates.OFF, wait=ANY)
    last_velocity_call = get_mock_put(smargon.omega.velocity).call_args_list[-1]
    assert last_velocity_call == call(initial_velocity, wait=ANY)


def test_given_thaw_succeeds_then_velocity_restored_and_thawer_turned_off(
    smargon: Smargon, thawer: Thawer, RE: RunEngine
):
    def do_thaw_func():
        RE(thaw(10, thawer=thawer, smargon=smargon))

    _do_thaw_and_confirm_cleanup(
        MagicMock(return_value=NullStatus()), smargon, thawer, do_thaw_func
    )


def test_given_moving_smargon_gives_error_then_velocity_restored_and_thawer_turned_off(
    smargon: Smargon, thawer: Thawer, RE: RunEngine
):
    def do_thaw_func():
        with pytest.raises(MyException):
            RE(thaw(10, thawer=thawer, smargon=smargon))

    _do_thaw_and_confirm_cleanup(
        MagicMock(side_effect=MyException()), smargon, thawer, do_thaw_func
    )


@pytest.mark.parametrize(
    "time, rotation, expected_speed",
    [
        (10, 360, 72),
        (3.5, 100, pytest.approx(57.142857)),
        (50, -100, 4),
    ],
)
def test_given_different_rotations_and_times_then_velocity_correct(
    smargon: Smargon,
    thawer: Thawer,
    time: float,
    rotation: float,
    expected_speed: float,
    RE: RunEngine,
):
    RE(thaw(time, rotation, thawer=thawer, smargon=smargon))
    first_velocity_call = get_mock_put(smargon.omega.velocity).call_args_list[0]
    assert first_velocity_call == call(expected_speed, wait=ANY)


@pytest.mark.parametrize(
    "start_pos, rotation, expected_end",
    [
        (0, 360, 360),
        (78, 100, 178),
        (0, -100, -100),
    ],
)
def test_given_different_rotations_then_motor_moved_relative(
    smargon: Smargon,
    thawer: Thawer,
    start_pos: float,
    rotation: float,
    expected_end: float,
    RE: RunEngine,
):
    set_mock_value(smargon.omega.user_setpoint, start_pos)
    RE(thaw(10, rotation, thawer=thawer, smargon=smargon))
    assert get_mock_put(smargon.omega.user_setpoint).call_args_list == [
        call(expected_end, wait=ANY),
        call(start_pos, wait=ANY),
    ]


@patch("mx_bluesky.beamlines.i04.thawing_plan.MurkoCallback")
async def test_thaw_and_stream_sets_sample_id_and_kicks_off_forwarder(
    patch_murko_callback: MagicMock,
    smargon: Smargon,
    thawer: Thawer,
    oav_forwarder: OAVToRedisForwarder,
    oav: OAV,
    robot: BartRobot,
    RE: RunEngine,
):
    set_mock_value(robot.sample_id, 100)
    RE(
        thaw_and_stream_to_redis(
            10,
            360,
            thawer=thawer,
            smargon=smargon,
            oav=oav,
            robot=robot,
            oav_to_redis_forwarder=oav_forwarder,
        )
    )
    assert await oav_forwarder.sample_id.get_value() == 100
    oav_forwarder.kickoff.assert_called()  # type: ignore
    oav_forwarder.complete.assert_called()  # type: ignore


@patch("mx_bluesky.beamlines.i04.thawing_plan.MurkoCallback")
def test_thaw_and_stream_adds_murko_callback_and_produces_expected_messages(
    patch_murko_callback: MagicMock,
    smargon: Smargon,
    thawer: Thawer,
    robot: BartRobot,
    oav_forwarder: OAVToRedisForwarder,
    oav: OAV,
    RE: RunEngine,
):
    patch_murko_instance = patch_murko_callback.return_value
    RE(
        thaw_and_stream_to_redis(
            10,
            360,
            thawer=thawer,
            smargon=smargon,
            oav=oav,
            robot=robot,
            oav_to_redis_forwarder=oav_forwarder,
        )
    )

    docs = patch_murko_instance.call_args_list
    start_params = [c.args[1] for c in docs if c.args[0] == "start"]
    event_params = [c.args[1] for c in docs if c.args[0] == "event"]
    assert len(start_params) == 1
    assert len(event_params) == 4
    oav_updates = [
        e for e in event_params if "oav_to_redis_forwarder-uuid" in e["data"].keys()
    ]
    smargon_updates = [e for e in event_params if "smargon-omega" in e["data"].keys()]
    assert len(oav_updates) > 0
    assert len(smargon_updates) > 0


@patch("mx_bluesky.beamlines.i04.thawing_plan.MurkoCallback.call_murko")
def test_thaw_and_stream_will_produce_events_that_call_murko(
    patch_murko_call: MagicMock,
    smargon: Smargon,
    thawer: Thawer,
    robot: BartRobot,
    oav_forwarder: OAVToRedisForwarder,
    oav: OAV,
    RE: RunEngine,
):
    RE(
        thaw_and_stream_to_redis(
            10,
            360,
            thawer=thawer,
            smargon=smargon,
            oav=oav,
            robot=robot,
            oav_to_redis_forwarder=oav_forwarder,
        )
    )
    patch_murko_call.assert_called()


def test_thaw_and_stream_will_switch_murko_source_half_way_through_thaw(
    sim_run_engine,
    smargon: Smargon,
    thawer: Thawer,
    oav_forwarder: OAVToRedisForwarder,
    oav: OAV,
    robot: BartRobot,
):
    msgs = sim_run_engine.simulate_plan(
        thaw_and_stream_to_redis(10, 360, robot, thawer, smargon, oav, oav_forwarder)
    )
    for source in [Source.FULL_SCREEN.value, Source.ROI.value]:
        msgs = assert_message_and_return_remaining(
            msgs,
            lambda msg: msg.command == "set"
            and msg.obj.name == "oav_to_redis_forwarder-selected_source"
            and msg.args[0] == source,
        )
        msgs = assert_message_and_return_remaining(
            msgs,
            lambda msg: msg.command == "kickoff"
            and msg.obj.name == "oav_to_redis_forwarder",
        )
        msgs = assert_message_and_return_remaining(
            msgs,
            lambda msg: msg.command == "set" and msg.obj.name == "smargon-omega",
        )
        msgs = assert_message_and_return_remaining(
            msgs,
            lambda msg: msg.command == "complete"
            and msg.obj.name == "oav_to_redis_forwarder",
        )


def _run_thaw_and_stream_and_assert_zoom_changes(
    smargon: Smargon,
    thawer: Thawer,
    oav_forwarder: OAVToRedisForwarder,
    oav: OAV,
    robot: BartRobot,
    RE: RunEngine,
    expect_raises=None,
):
    set_mock_value(oav.zoom_controller.level, "2.0x")

    run_plan = partial(
        RE,
        thaw_and_stream_to_redis(
            10,
            360,
            thawer=thawer,
            smargon=smargon,
            oav=oav,
            robot=robot,
            oav_to_redis_forwarder=oav_forwarder,
        ),
    )

    if expect_raises:
        with pytest.raises(expect_raises):
            run_plan()
    else:
        run_plan()

    mock_level_set = get_mock_put(oav.zoom_controller.level)
    mock_level_set.assert_has_calls([call("1.0x", wait=True), call("2.0x", wait=True)])


@patch("mx_bluesky.beamlines.i04.thawing_plan.MurkoCallback")
def test_given_thaw_succeeds_then_thaw_and_stream_sets_zoom_to_1_and_back(
    patch_murko_callback,
    smargon: Smargon,
    thawer: Thawer,
    oav_forwarder: OAVToRedisForwarder,
    oav: OAV,
    robot: BartRobot,
    RE: RunEngine,
):
    _run_thaw_and_stream_and_assert_zoom_changes(
        smargon, thawer, oav_forwarder, oav, robot, RE
    )


@patch("mx_bluesky.beamlines.i04.thawing_plan.MurkoCallback")
@patch("mx_bluesky.beamlines.i04.thawing_plan._thaw")
def test_given_thaw_fails_then_thaw_and_stream_sets_zoom_to_1_and_back(
    mock__thaw,
    patch_murko_callback,
    smargon: Smargon,
    thawer: Thawer,
    oav_forwarder: OAVToRedisForwarder,
    oav: OAV,
    robot: BartRobot,
    RE: RunEngine,
):
    mock__thaw.side_effect = Exception()
    _run_thaw_and_stream_and_assert_zoom_changes(
        smargon, thawer, oav_forwarder, oav, robot, RE, Exception
    )



from mx_bluesky.beamlines.aithre_lasershaping.check_goniometer_performance import (
    check_omega_performance,
)
from mx_bluesky.beamlines.aithre_lasershaping.goniometer_controls import (
    change_goniometer_turn_speed,
    rotate_goniometer_relative,
)

__all__ = [
    "check_omega_performance",
    "change_goniometer_turn_speed",
    "rotate_goniometer_relative",
]



import io
from functools import partial
from unittest.mock import MagicMock, patch

import pin_versions
import pytest


@pytest.fixture
def patched_run_pip_freeze():
    with patch("pin_versions.run_pip_freeze") as run_pip_freeze:
        with open("test_data/pip_freeze.txt") as freeze_output:
            mock_process = MagicMock()
            run_pip_freeze.return_value = mock_process
            mock_process.stdout = freeze_output.read()
            mock_process.returncode = 0
        yield run_pip_freeze


@pytest.mark.parametrize(
    "input, expected",
    [
        ("", ("", None)),
        (
            "    pydantic<2.0 # See https://github.com/DiamondLightSource/hyperion/issues/774",
            (
                "    pydantic<2.0 ",
                " See https://github.com/DiamondLightSource/hyperion/issues/774",
            ),
        ),
    ],
)
def test_strip_comment(input, expected):
    assert pin_versions.strip_comment(input) == expected


@pytest.mark.parametrize(
    "input, expected",
    [
        ("dls-dodal", "dls-dodal"),
        ("dls_dodal", "dls-dodal"),
        ("dls.dodal", "dls-dodal"),
    ],
)
def test_normalize(input, expected):
    assert pin_versions.normalize(input) == expected


def test_unpin():
    with io.StringIO() as output_file:
        with open("test_data/setup.cfg") as input_file:
            pin_versions.process_files(
                input_file, output_file, pin_versions.unpin_versions
            )
        with open("test_data/setup.cfg.unpinned") as expected_file:
            assert output_file.getvalue() == expected_file.read()


@patch("pin_versions.stdout")
def test_write_commit_message(mock_stdout, patched_run_pip_freeze):
    installed_versions = pin_versions.fetch_pin_versions()
    pin_versions.write_commit_message(installed_versions)
    mock_stdout.write.assert_called_once_with(
        "Pin dependencies prior to release. Dodal 1.13.1, nexgen 0.8.4"
    )


def test_pin(patched_run_pip_freeze):
    installed_versions = pin_versions.fetch_pin_versions()
    with io.StringIO() as output_file:
        with open("test_data/setup.cfg.unpinned") as input_file:
            pin_versions.process_files(
                input_file,
                output_file,
                partial(pin_versions.update_setup_cfg_line, installed_versions),
            )

        with open("test_data/setup.cfg.pinned") as expected_file:
            assert output_file.getvalue() == expected_file.read()



#!/usr/bin/env python3
from inspect import get_annotations, getmodule, getmro, isclass
from typing import get_type_hints

from blueapi.core.bluesky_types import is_bluesky_plan_generator
from blueapi.core.context import load_module_all
from pydantic import BaseModel

from mx_bluesky.common.parameters.components import MxBlueskyParameters
from mx_bluesky.hyperion import experiment_plans


def main():
    """
    Generate the PlantUML source for a diagram of the parameter models on the standard output.
    """
    print("""
'This file is auto-generated by generate_plantuml.py
@startuml hyperion_parameter_model
title Hyperion Parameter Model
set namespaceSeparator none
""")

    parameter_types = [
        get_type_hints(obj)["parameters"]
        for obj in load_module_all(experiment_plans)
        if is_bluesky_plan_generator(obj)
    ]
    experiment_types = set()
    all_types = set()
    for parameter_type in parameter_types:
        experiment_types.add(parameter_type)
        all_types.add(parameter_type)

    for experiment_type in experiment_types:
        for base in getmro(experiment_type):
            if issubclass(base, BaseModel) and base is not BaseModel:
                all_types.add(base)

    types_by_package = {}
    for t in all_types:
        mod = getmodule(t)
        assert mod
        types_by_package.setdefault(mod.__package__, []).append(t)

    mx_bluesky_param_types = set()
    mixin_types = set()

    for t in types_by_package["mx_bluesky.common.parameters"]:
        if issubclass(t, MxBlueskyParameters):
            mx_bluesky_param_types.add(t)
        else:
            mixin_types.add(t)

    print("package mx_bluesky.common.parameters {")
    print("together {")
    for t in mixin_types:
        generate_class(t)
    print("}")
    print("together {")
    for t in mx_bluesky_param_types:
        generate_class(t)
    print("}")
    print("}")

    print("package mx_bluesky.hyperion.parameters {")
    for t in types_by_package["mx_bluesky.hyperion.parameters"]:
        generate_class(t)
    print("}")

    for t in all_types:
        for base in t.__bases__:
            if base is not object and base is not BaseModel:
                print(f"{base.__name__} <|-- {t.__name__}")

    print("@enduml")


def generate_class(t):
    print(f"class {t.__name__}{{")
    for field_name, field_type in get_annotations(t).items():
        print(f"\t{generate_type(field_type)} {field_name}")
    print("}")


def generate_type(field_type):
    return (
        field_type.__name__
        if isclass(field_type) and issubclass(field_type, BaseModel)
        else str(field_type)
    )


if __name__ == "__main__":
    main()



import copy
import json
from datetime import timedelta

from bluesky.callbacks import CallbackBase
from dodal.log import LOGGER
from event_model.documents import Event, RunStart, RunStop
from redis import StrictRedis


class MurkoCallback(CallbackBase):
    DATA_EXPIRY_DAYS = 7

    def __init__(self, redis_host: str, redis_password: str, redis_db: int = 0):
        self.redis_client = StrictRedis(
            host=redis_host, password=redis_password, db=redis_db
        )
        self.last_uuid = None

    def start(self, doc: RunStart) -> RunStart | None:
        self.sample_id = doc.get("sample_id")
        self.murko_metadata = {
            "zoom_percentage": doc.get("zoom_percentage"),
            "microns_per_x_pixel": doc.get("microns_per_x_pixel"),
            "microns_per_y_pixel": doc.get("microns_per_y_pixel"),
            "beam_centre_i": doc.get("beam_centre_i"),
            "beam_centre_j": doc.get("beam_centre_j"),
            "sample_id": self.sample_id,
        }
        self.last_uuid = None
        LOGGER.info(f"Starting to stream metadata to murko under {self.sample_id}")
        return doc

    def event(self, doc: Event) -> Event:
        if latest_omega := doc["data"].get("smargon-omega"):
            if self.last_uuid is not None:
                self.call_murko(self.last_uuid, latest_omega)
        elif (uuid := doc["data"].get("oav_to_redis_forwarder-uuid")) is not None:
            self.last_uuid = uuid
        return doc

    def call_murko(self, uuid: str, omega: float):
        metadata = copy.deepcopy(self.murko_metadata)
        metadata["omega_angle"] = omega
        metadata["uuid"] = uuid

        # Send metadata to REDIS and trigger murko
        redis_key = f"murko:{metadata['sample_id']}:metadata"
        self.redis_client.hset(redis_key, uuid, json.dumps(metadata))
        self.redis_client.expire(redis_key, timedelta(days=self.DATA_EXPIRY_DAYS))
        self.redis_client.publish("murko", json.dumps(metadata))

    def stop(self, doc: RunStop) -> RunStop | None:
        LOGGER.info(f"Finished streaming {self.sample_id} to murko")
        return doc



import os
from pathlib import Path
from shutil import copy
from unittest.mock import patch

import bluesky.preprocessors as bpp
import h5py
import numpy as np
import pytest
from bluesky.run_engine import RunEngine
from h5py import Dataset, ExternalLink, Group

from mx_bluesky.common.external_interaction.nexus.write_nexus import NexusWriter
from mx_bluesky.common.plans.read_hardware import (
    standard_read_hardware_during_collection,
)
from mx_bluesky.common.utils.log import LOGGER
from mx_bluesky.hyperion.experiment_plans.rotation_scan_plan import (
    RotationScanComposite,
)
from mx_bluesky.hyperion.external_interaction.callbacks.rotation.nexus_callback import (
    RotationNexusFileCallback,
)
from mx_bluesky.hyperion.parameters.constants import CONST
from mx_bluesky.hyperion.parameters.rotation import RotationScan

from ....conftest import extract_metafile, raw_params_from_file

TEST_EXAMPLE_NEXUS_FILE = Path("ins_8_5.nxs")
TEST_EXAMPLE_NEXUS_METAFILE_PREFIX = "ins_8_5_meta"
TEST_DATA_DIRECTORY = Path("tests/test_data/nexus_files/rotation")
TEST_FILENAME = "rotation_scan_test_nexus"


@pytest.fixture
def test_params(tmpdir):
    param_dict = raw_params_from_file(
        "tests/test_data/parameter_json_files/good_test_rotation_scan_parameters.json"
    )
    param_dict["storage_directory"] = "tests/test_data"
    param_dict["file_name"] = f"{tmpdir}/{TEST_FILENAME}"
    param_dict["scan_width_deg"] = 360.0
    param_dict["demand_energy_ev"] = 12700
    params = RotationScan(**param_dict)
    params.x_start_um = 0
    params.y_start_um = 0
    params.z_start_um = 0
    params.exposure_time_s = 0.004
    return params


def fake_rotation_scan(
    parameters: RotationScan,
    subscription: RotationNexusFileCallback,
    rotation_devices: RotationScanComposite,
):
    @bpp.subs_decorator(subscription)
    @bpp.set_run_key_decorator("rotation_scan_with_cleanup_and_subs")
    @bpp.run_decorator(  # attach experiment metadata to the start document
        md={
            "subplan_name": CONST.PLAN.ROTATION_OUTER,
            "mx_bluesky_parameters": parameters.model_dump_json(),
            "activate_callbacks": "RotationNexusFileCallback",
        }
    )
    def plan():
        yield from standard_read_hardware_during_collection(
            rotation_devices.aperture_scatterguard,
            rotation_devices.attenuator,
            rotation_devices.flux,
            rotation_devices.dcm,
            rotation_devices.eiger,
        )

    return plan()


def dectris_device_mapping(meta_filename: str):
    return {
        "entry": {
            "instrument": {
                "detector": {
                    "bit_depth_image": f"{meta_filename}//_dectris/bit_depth_image",
                    "bit_depth_readout": f"{meta_filename}//_dectris/bit_depth_image",
                    "detectorSpecific": {
                        "ntrigger": f"{meta_filename}///_dectris/ntrigger",
                        "software_version": f"{meta_filename}//_dectris/software_version",
                    },
                    "detector_readout_time": f"{meta_filename}//_dectris/detector_readout_time",
                    "flatfield_applied": f"{meta_filename}//_dectris/flatfield_correction_applied",
                    "photon_energy": f"{meta_filename}//_dectris/photon_energy",
                    "pixel_mask": f"{meta_filename}//mask",
                    "pixel_mask_applied": f"{meta_filename}//_dectris/pixel_mask_applied",
                    "threshold_energy": f"{meta_filename}//_dectris/threshold_energy",
                }
            }
        }
    }


def apply_metafile_mapping(exceptions: dict, mapping: dict):
    """Recursively populate the exceptions map with corresponding mapping entries"""
    for key in mapping.keys():
        mapping_value = mapping.get(key)
        if isinstance(mapping_value, dict):
            exceptions_child = exceptions.setdefault(key, {})
            apply_metafile_mapping(exceptions_child, mapping_value)
        else:
            exceptions[key] = mapping_value


def test_rotation_scan_nexus_output_compared_to_existing_full_compare(
    test_params: RotationScan,
    tmpdir,
    fake_create_rotation_devices: RotationScanComposite,
):
    test_params.chi_start_deg = 0
    test_params.phi_start_deg = 0
    run_number = test_params.detector_params.run_number
    nexus_filename = f"{tmpdir}/{TEST_FILENAME}_{run_number}.nxs"
    master_filename = f"{tmpdir}/{TEST_FILENAME}_{run_number}_master.h5"
    meta_filename = f"{TEST_FILENAME}_{run_number}_meta.h5"

    fake_create_rotation_devices.eiger.bit_depth.sim_put(32)  # type: ignore

    RE = RunEngine({})

    with patch(
        "mx_bluesky.common.external_interaction.nexus.write_nexus.get_start_and_predicted_end_time",
        return_value=("test_time", "test_time"),
    ):
        RE(
            fake_rotation_scan(
                test_params, RotationNexusFileCallback(), fake_create_rotation_devices
            )
        )

    assert os.path.isfile(nexus_filename)
    assert os.path.isfile(master_filename)

    example_metafile_path = (
        f"{TEST_DATA_DIRECTORY}/{TEST_EXAMPLE_NEXUS_METAFILE_PREFIX}.h5.gz"
    )
    extract_metafile(
        example_metafile_path, f"{tmpdir}/{TEST_EXAMPLE_NEXUS_METAFILE_PREFIX}.h5"
    )
    example_nexus_path = f"{tmpdir}/{TEST_EXAMPLE_NEXUS_FILE}"
    copy(TEST_DATA_DIRECTORY / TEST_EXAMPLE_NEXUS_FILE, example_nexus_path)

    # Models expected differences to the GDA master nexus file
    # If a key is in _missing then it is not expected to be present
    # If a key is in _ignore then we do not compare it
    # If a key maps to a dict then we expect it to be a Group
    # Otherwise if a key is present we expect it to be a DataSet
    # If a key maps to a callable then we use that as the comparison function
    # Otherwise we compare the scalar or array value as appropriate
    exceptions = {
        "entry": {
            "_missing": {"end_time"},
            "data": {"_ignore": {"data", "omega"}},
            "instrument": {
                "_missing": {"transformations", "detector_z", "source"},
                "detector": {
                    "_missing": {
                        "detector_distance",
                        "serial_number",  # nexgen#236
                    },
                    "distance": 0.1,
                    "transformations": {
                        "detector_z": {"det_z": np.array([100])},
                        "det_z": np.array([100]),
                    },
                    "detector_z": {"det_z": np.array([100])},
                    "underload_value": 0,
                    "_ignore": {
                        "beam_center_x",
                        "beam_center_y",
                        "depends_on",
                        "saturation_value",
                        "sensor_material",
                    },
                    "detectorSpecific": {
                        "_missing": {"pixel_mask"},
                        "x_pixels": 4148,
                        "y_pixels": 4362,
                    },
                    "module": {"_ignore": {"module_offset"}},
                    "sensor_thickness": np.isclose,
                },
                "attenuator": {"attenuator_transmission": np.isclose},
                "beam": {"incident_wavelength": np.isclose},
                "name": b"DIAMOND BEAMLINE S03",
            },
            "sample": {
                "beam": {"incident_wavelength": np.isclose},
                "transformations": {
                    "_missing": {"omega_end"},
                    "_ignore": {"omega"},
                    "omega_increment_set": 0.1,
                    "omega_end": lambda a, b: np.all(np.isclose(a, b, atol=1e-03)),
                },
                "sample_omega": {
                    "_ignore": {"omega_end", "omega"},
                    "omega_increment_set": 0.1,
                },
                "sample_x": {"sam_x": np.isclose},
                "sample_y": {"sam_y": np.isclose},
                "sample_z": {"sam_z": np.isclose},
                "sample_chi": {"chi": np.isclose},
                "sample_phi": {"phi": np.isclose},
            },
            "end_time_estimated": b"test_timeZ",
            "start_time": b"test_timeZ",
            "source": {
                "name": b"Diamond Light Source",
                "type": b"Synchrotron X-ray Source",
            },
        }
    }

    with (
        h5py.File(example_nexus_path, "r") as example_nexus,
        h5py.File(nexus_filename, "r") as hyperion_nexus,
    ):
        apply_metafile_mapping(exceptions, dectris_device_mapping(meta_filename))
        _compare_actual_and_expected_nexus_output(
            hyperion_nexus, example_nexus, exceptions
        )


def test_rotation_scan_nexus_output_compared_to_existing_file(
    test_params: RotationScan,
    tmpdir,
    fake_create_rotation_devices: RotationScanComposite,
):
    run_number = test_params.run_number or test_params.detector_params.run_number
    nexus_filename = f"{tmpdir}/{TEST_FILENAME}_{run_number}.nxs"
    master_filename = f"{tmpdir}/{TEST_FILENAME}_{run_number}_master.h5"

    fake_create_rotation_devices.eiger.bit_depth.sim_put(32)  # type: ignore

    RE = RunEngine({})

    with patch(
        "mx_bluesky.common.external_interaction.nexus.write_nexus.get_start_and_predicted_end_time",
        return_value=("test_time", "test_time"),
    ):
        RE(
            fake_rotation_scan(
                test_params, RotationNexusFileCallback(), fake_create_rotation_devices
            )
        )

    assert os.path.isfile(nexus_filename)
    assert os.path.isfile(master_filename)

    with (
        h5py.File(
            str(TEST_DATA_DIRECTORY / TEST_EXAMPLE_NEXUS_FILE), "r"
        ) as example_nexus,
        h5py.File(nexus_filename, "r") as hyperion_nexus,
    ):
        assert hyperion_nexus["/entry/start_time"][()] == b"test_timeZ"  # type: ignore
        assert hyperion_nexus["/entry/end_time_estimated"][()] == b"test_timeZ"  # type: ignore

        # we used to write the positions wrong...
        hyperion_omega: np.ndarray = np.array(
            hyperion_nexus["/entry/data/omega"][:]  # type: ignore
        )
        example_omega: np.ndarray = example_nexus["/entry/data/omega"][:]  # type: ignore
        assert np.allclose(hyperion_omega, example_omega)

        assert isinstance(
            hyperion_data := hyperion_nexus["/entry/data/data"], h5py.Dataset
        )
        example_data_shape = example_nexus["/entry/data/data"].shape  # type: ignore

        assert hyperion_data.dtype == "uint32"
        assert hyperion_data.shape == example_data_shape

        hyperion_instrument = hyperion_nexus["/entry/instrument"]
        example_instrument = example_nexus["/entry/instrument"]
        transmission = "attenuator/attenuator_transmission"
        wavelength = "beam/incident_wavelength"
        assert np.isclose(
            hyperion_instrument[transmission][()],  # type: ignore
            example_instrument[transmission][()],  # type: ignore
        )
        assert np.isclose(
            hyperion_instrument[wavelength][()],  # type: ignore
            example_instrument[wavelength][()],  # type: ignore
        )

        hyperion_sam_x = hyperion_nexus["/entry/sample/sample_x/sam_x"]
        example_sam_x = example_nexus["/entry/sample/sample_x/sam_x"]
        hyperion_sam_y = hyperion_nexus["/entry/sample/sample_y/sam_y"]
        example_sam_y = example_nexus["/entry/sample/sample_y/sam_y"]
        hyperion_sam_z = hyperion_nexus["/entry/sample/sample_z/sam_z"]
        example_sam_z = example_nexus["/entry/sample/sample_z/sam_z"]

        hyperion_sam_phi = hyperion_nexus["/entry/sample/sample_phi/phi"]
        example_sam_phi = example_nexus["/entry/sample/sample_phi/phi"]
        hyperion_sam_chi = hyperion_nexus["/entry/sample/sample_chi/chi"]
        example_sam_chi = example_nexus["/entry/sample/sample_chi/chi"]

        hyperion_sam_omega = hyperion_nexus["/entry/sample/sample_omega/omega"]
        example_sam_omega = example_nexus["/entry/sample/sample_omega/omega"]

        assert np.isclose(
            hyperion_sam_x[()],  # type: ignore
            example_sam_x[()],  # type: ignore
        )
        assert np.isclose(
            hyperion_sam_y[()],  # type: ignore
            example_sam_y[()],  # type: ignore
        )
        assert np.isclose(
            hyperion_sam_z[()],  # type: ignore
            example_sam_z[()],  # type: ignore
        )

        assert hyperion_sam_x.attrs.get("depends_on") == example_sam_x.attrs.get(
            "depends_on"
        )
        assert hyperion_sam_y.attrs.get("depends_on") == example_sam_y.attrs.get(
            "depends_on"
        )
        assert hyperion_sam_z.attrs.get("depends_on") == example_sam_z.attrs.get(
            "depends_on"
        )
        assert hyperion_sam_phi.attrs.get("depends_on") == example_sam_phi.attrs.get(
            "depends_on"
        )
        assert hyperion_sam_chi.attrs.get("depends_on") == example_sam_chi.attrs.get(
            "depends_on"
        )
        assert hyperion_sam_omega.attrs.get(
            "depends_on"
        ) == example_sam_omega.attrs.get("depends_on")


@pytest.mark.parametrize(
    "bit_depth,expected_type",
    [(8, np.uint8), (16, np.uint16), (32, np.uint32), (100, np.uint16)],
)
@patch("mx_bluesky.common.external_interaction.nexus.write_nexus.NXmxFileWriter")
def test_given_detector_bit_depth_changes_then_vds_datatype_as_expected(
    mock_nexus_writer,
    test_params: RotationScan,
    fake_create_rotation_devices: RotationScanComposite,
    bit_depth,
    expected_type,
):
    write_vds_mock = mock_nexus_writer.return_value.write_vds

    fake_create_rotation_devices.eiger.bit_depth.sim_put(bit_depth)  # type: ignore

    RE = RunEngine({})

    with patch(
        "mx_bluesky.common.external_interaction.nexus.write_nexus.get_start_and_predicted_end_time",
        return_value=("test_time", "test_time"),
    ):
        RE(
            fake_rotation_scan(
                test_params, RotationNexusFileCallback(), fake_create_rotation_devices
            )
        )

        for call in write_vds_mock.mock_calls:
            assert call.kwargs["vds_dtype"] == expected_type


def _compare_actual_and_expected_nexus_output(actual, expected, exceptions: dict):
    _compare_actual_and_expected([], actual, expected, exceptions)


def _compare_actual_and_expected(path: list[str], actual, expected, exceptions: dict):
    if expected is None:
        # The nexus file under test contains a node that isn't in the original GDA reference nexus file
        # but we may still expect something if the exception map contains it
        expected = {}

    path_str = "/".join(path)
    LOGGER.debug(f"Comparing {path_str}")
    keys_not_in_actual = (
        expected.keys() - actual.keys() - exceptions.get("_missing", set())
    )
    assert len(keys_not_in_actual) == 0, (
        f"Missing entries in group {path_str}, {keys_not_in_actual}"
    )

    keys_to_compare = actual.keys()
    keys_to_ignore = exceptions.get("_ignore", set())
    keys_not_in_expected = (
        keys_to_compare
        - expected.keys()
        - {k for k in exceptions.keys() if not k.startswith("_")}
        - keys_to_ignore
    )
    cmp = len(keys_not_in_expected) == 0
    keys_to_compare = sorted(keys_to_compare)
    assert cmp, f"Found unexpected entries in group {path_str}, {keys_not_in_expected}"
    for key in keys_to_compare:
        item_path = path + [key]
        actual_link = actual.get(key, getlink=True)
        item_path_str = "/" + "/".join(item_path)
        exception = exceptions.get(key, None)
        if isinstance(actual_link, ExternalLink):
            if exception:
                actual_link_path = f"{actual_link.filename}//{actual_link.path}"
                assert actual_link_path == exception, (
                    f"Actual and expected external links differ {actual_link_path}, {exception}"
                )
            else:
                LOGGER.debug(
                    f"Skipping external link {item_path_str} -> {actual_link.path}"
                )
            continue
        actual_class = actual.get(key, getclass=True, getlink=False)
        expected_class = (
            Group
            if isinstance(exception, dict)
            else (
                Dataset if exception is not None else expected.get(key, getclass=True)  # type: ignore
            )
        )
        actual_value = actual.get(key)
        expected_value = (
            expected.get(key)
            if (exception is None or isinstance(exception, dict) or callable(exception))
            else exception
        )
        if expected_class == Group:
            _compare_actual_and_expected(
                item_path, actual_value, expected.get(key), exceptions.get(key, {})
            )
        elif (expected_class == Dataset) and key not in keys_to_ignore:
            if isinstance(expected_value, Dataset):
                # Only check shape if we didn't override the expected value
                assert actual_value.shape == expected_value.shape, (
                    f"Actual and expected shapes differ for {item_path_str}: {actual_value.shape}, {expected_value.shape}"
                )
            else:
                assert hasattr(actual_value, "shape"), f"No shape for {item_path_str}"
                expected_shape = np.shape(expected_value)  # type: ignore
                assert actual_value.shape == expected_shape, (
                    f"{item_path_str} data shape not expected shape{actual_value.shape}, {expected_shape}"
                )
            if actual_value.shape == ():
                if callable(exception):
                    assert exceptions.get(key)(actual_value, expected_value)  # type: ignore
                elif np.isscalar(exception):
                    assert actual_value[()] == exception, (
                        f"{item_path_str} actual and expected did not match {actual_value[()]}, {exception}."
                    )
                else:
                    assert actual_class == expected_class, (
                        f"{item_path_str} Actual and expected class don't match {actual_class}, {expected_class}"
                    )
                    # fmt: off
                    assert (
                        actual_value[()] == expected_value[()]  # type: ignore
                    ), (
                        f"Actual and expected values differ for {item_path_str}: "
                        f"{actual_value[()]} != {expected_value[()]}"  # type: ignore
                    )
                    # fmt: on
            else:
                actual_value_str = np.array2string(actual_value, threshold=10)
                expected_value_str = np.array2string(expected_value, threshold=10)  # type: ignore
                if callable(exception):
                    assert exception(actual_value, expected_value), (
                        f"Actual and expected values differ for {item_path_str}: {actual_value_str} != {expected_value_str}, according to {exception}"
                    )
                else:
                    assert np.array_equal(
                        actual_value,
                        expected_value,  # type: ignore
                    ), (
                        f"Actual and expected values differ for {item_path_str}: {actual_value_str} != {expected_value_str}"
                    )


def test_override_parameters_override(test_params: RotationScan):
    writer = NexusWriter(
        test_params, (1, 2, 3), {}, full_num_of_images=82367, meta_data_run_number=9852
    )
    assert writer.full_num_of_images != test_params.num_images
    assert writer.full_num_of_images == 82367
    assert writer.data_filename == f"{test_params.file_name}_9852"



from __future__ import annotations

from typing import TYPE_CHECKING

from bluesky.callbacks import CallbackBase
from dodal.devices.zocalo import ZocaloStartInfo, ZocaloTrigger

from mx_bluesky.common.parameters.constants import (
    DocDescriptorNames,
)
from mx_bluesky.common.utils.exceptions import ISPyBDepositionNotMade
from mx_bluesky.common.utils.log import ISPYB_ZOCALO_CALLBACK_LOGGER
from mx_bluesky.common.utils.utils import number_of_frames_from_scan_spec

if TYPE_CHECKING:
    from event_model.documents import Event, EventDescriptor, RunStart, RunStop


class ZocaloCallback(CallbackBase):
    """Callback class to handle the triggering of Zocalo processing.
    Will start listening for collections when {triggering_plan} has been started.

    For every ispyb deposition that occurs inside this run the callback will send zocalo
    a run_start signal. Once the {triggering_plan} has ended the callback will send a
    run_end signal for all collections.

    Shouldn't be subscribed directly to the RunEngine, instead should be passed to the
    `emit` argument of an ISPyB callback which appends DCIDs to the relevant start doc.
    """

    def _reset_state(self):
        self.run_uid: str | None = None
        self.zocalo_info: list[ZocaloStartInfo] = []
        self._started_zocalo_collections: list[ZocaloStartInfo] = []
        self.descriptors: dict[str, EventDescriptor] = {}
        self.start_frame = 0

    def __init__(self, triggering_plan: str, zocalo_environment: str):
        super().__init__()
        self.triggering_plan = triggering_plan
        self.zocalo_interactor = ZocaloTrigger(zocalo_environment)
        self._reset_state()

    def start(self, doc: RunStart):
        ISPYB_ZOCALO_CALLBACK_LOGGER.info("Zocalo handler received start document.")
        if self.triggering_plan and doc.get("subplan_name") == self.triggering_plan:
            self.run_uid = doc.get("uid")
        if self.run_uid:
            if (
                isinstance(scan_points := doc.get("scan_points"), list)
                and isinstance(ispyb_ids := doc.get("ispyb_dcids"), tuple)
                and len(ispyb_ids) > 0
            ):
                ISPYB_ZOCALO_CALLBACK_LOGGER.info(f"Zocalo triggering for {ispyb_ids}")
                ids_and_shape = list(zip(ispyb_ids, scan_points, strict=False))
                for idx, id_and_shape in enumerate(ids_and_shape):
                    id, shape = id_and_shape
                    num_frames = number_of_frames_from_scan_spec(shape)
                    self.zocalo_info.append(
                        ZocaloStartInfo(id, None, self.start_frame, num_frames, idx)
                    )
                    self.start_frame += num_frames

    def descriptor(self, doc: EventDescriptor):
        self.descriptors[doc["uid"]] = doc

    def event(self, doc: Event) -> Event:
        event_descriptor = self.descriptors[doc["descriptor"]]
        if event_descriptor.get("name") == DocDescriptorNames.ZOCALO_HW_READ:
            filename = doc["data"]["eiger_odin_file_writer_id"]
            for start_info in self.zocalo_info:
                start_info.filename = filename
                self.zocalo_interactor.run_start(start_info)
                self._started_zocalo_collections.append(start_info)
            self.zocalo_info = []
        return doc

    def stop(self, doc: RunStop):
        if doc.get("run_start") == self.run_uid:
            ISPYB_ZOCALO_CALLBACK_LOGGER.info(
                f"Zocalo handler received stop document, for run {doc.get('run_start')}."
            )
            if not self._started_zocalo_collections:
                raise ISPyBDepositionNotMade(
                    f"No ISPyB IDs received by the end of {self.triggering_plan=}"
                )
            for info in self._started_zocalo_collections:
                self.zocalo_interactor.run_end(info.ispyb_dcid)
            self._reset_state()



from unittest.mock import MagicMock, patch

import pytest
from graypy import GELFTCPHandler

from mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_callback import (
    GridscanISPyBCallback,
)
from mx_bluesky.common.external_interaction.ispyb.ispyb_store import (
    IspybIds,
    StoreInIspyb,
)
from mx_bluesky.common.utils.log import ISPYB_ZOCALO_CALLBACK_LOGGER
from mx_bluesky.hyperion.external_interaction.callbacks.__main__ import setup_logging
from mx_bluesky.hyperion.parameters.gridscan import GridCommonWithHyperionDetectorParams

from .....conftest import TestData

DC_IDS = (1, 2)
DCG_ID = 4
DC_GRID_IDS = (11, 12)
td = TestData()


def mock_store_in_ispyb(config, *args, **kwargs) -> StoreInIspyb:
    mock = MagicMock(spec=StoreInIspyb)
    mock.end_deposition = MagicMock(return_value=None)
    mock.begin_deposition = MagicMock(
        return_value=IspybIds(
            data_collection_group_id=DCG_ID, data_collection_ids=DC_IDS
        )
    )
    mock.update_deposition = MagicMock(
        return_value=IspybIds(
            data_collection_group_id=DCG_ID,
            data_collection_ids=DC_IDS,
            grid_ids=DC_GRID_IDS,
        )
    )
    mock.append_to_comment = MagicMock()
    return mock


@patch(
    "mx_bluesky.common.external_interaction.callbacks.common.ispyb_mapping.get_current_time_string",
    MagicMock(return_value=td.DUMMY_TIME_STRING),
)
@patch(
    "mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_callback.StoreInIspyb",
    mock_store_in_ispyb,
)
class TestXrayCentreIspybHandler:
    def test_fgs_failing_results_in_bad_run_status_in_ispyb(
        self,
    ):
        ispyb_handler = GridscanISPyBCallback(
            param_type=GridCommonWithHyperionDetectorParams
        )
        ispyb_handler.activity_gated_start(td.test_gridscan3d_start_document)
        ispyb_handler.activity_gated_descriptor(
            td.test_descriptor_document_pre_data_collection
        )
        ispyb_handler.activity_gated_event(td.test_event_document_pre_data_collection)
        ispyb_handler.activity_gated_descriptor(
            td.test_descriptor_document_during_data_collection
        )
        ispyb_handler.activity_gated_event(
            td.test_event_document_during_data_collection  # pyright: ignore
        )
        ispyb_handler.activity_gated_stop(td.test_run_gridscan_failed_stop_document)

        ispyb_handler.ispyb.end_deposition.assert_called_once_with(  # type: ignore
            IspybIds(
                data_collection_group_id=DCG_ID,
                data_collection_ids=DC_IDS,
                grid_ids=DC_GRID_IDS,
            ),
            "fail",
            "could not connect to devices",
        )

    def test_fgs_raising_no_exception_results_in_good_run_status_in_ispyb(
        self,
    ):
        ispyb_handler = GridscanISPyBCallback(
            param_type=GridCommonWithHyperionDetectorParams
        )
        ispyb_handler.activity_gated_start(td.test_gridscan3d_start_document)
        ispyb_handler.activity_gated_descriptor(
            td.test_descriptor_document_pre_data_collection
        )
        ispyb_handler.activity_gated_event(td.test_event_document_pre_data_collection)
        ispyb_handler.activity_gated_descriptor(
            td.test_descriptor_document_during_data_collection
        )
        ispyb_handler.activity_gated_event(
            td.test_event_document_during_data_collection
        )
        ispyb_handler.activity_gated_stop(td.test_do_fgs_gridscan_stop_document)

        ispyb_handler.ispyb.end_deposition.assert_called_once_with(  # type: ignore
            IspybIds(
                data_collection_group_id=DCG_ID,
                data_collection_ids=DC_IDS,
                grid_ids=DC_GRID_IDS,
            ),
            "success",
            "",
        )

    @pytest.mark.skip_log_setup
    def test_given_ispyb_callback_started_writing_to_ispyb_when_messages_logged_then_they_contain_dcgid(
        self,
    ):
        setup_logging(True)
        gelf_handler: MagicMock = next(
            filter(
                lambda h: isinstance(h, GELFTCPHandler),
                ISPYB_ZOCALO_CALLBACK_LOGGER.handlers,  # type: ignore
            )
        )
        gelf_handler.emit = MagicMock()

        ispyb_handler = GridscanISPyBCallback(
            param_type=GridCommonWithHyperionDetectorParams
        )
        ispyb_handler.activity_gated_start(td.test_gridscan3d_start_document)
        ispyb_handler.activity_gated_descriptor(
            td.test_descriptor_document_pre_data_collection
        )
        ispyb_handler.activity_gated_event(td.test_event_document_pre_data_collection)
        ispyb_handler.activity_gated_descriptor(
            td.test_descriptor_document_during_data_collection
        )
        ispyb_handler.activity_gated_event(
            td.test_event_document_during_data_collection
        )

        ISPYB_ZOCALO_CALLBACK_LOGGER.info("test")
        latest_record = gelf_handler.emit.call_args.args[-1]
        assert latest_record.dc_group_id == DCG_ID

    @pytest.mark.skip_log_setup
    def test_given_ispyb_callback_finished_writing_to_ispyb_when_messages_logged_then_they_do_not_contain_dcgid(
        self,
    ):
        setup_logging(True)
        gelf_handler: MagicMock = next(
            filter(
                lambda h: isinstance(h, GELFTCPHandler),
                ISPYB_ZOCALO_CALLBACK_LOGGER.handlers,  # type: ignore
            )
        )
        gelf_handler.emit = MagicMock()

        ispyb_handler = GridscanISPyBCallback(
            param_type=GridCommonWithHyperionDetectorParams
        )
        ispyb_handler.activity_gated_start(td.test_gridscan3d_start_document)
        ispyb_handler.activity_gated_descriptor(
            td.test_descriptor_document_pre_data_collection
        )
        ispyb_handler.activity_gated_event(td.test_event_document_pre_data_collection)
        ispyb_handler.activity_gated_descriptor(
            td.test_descriptor_document_during_data_collection
        )
        ispyb_handler.activity_gated_event(
            td.test_event_document_during_data_collection
        )
        ispyb_handler.activity_gated_stop(td.test_run_gridscan_failed_stop_document)

        ISPYB_ZOCALO_CALLBACK_LOGGER.info("test")
        latest_record = gelf_handler.emit.call_args.args[-1]
        assert not hasattr(latest_record, "dc_group_id")

    @patch(
        "mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_callback.time",
        side_effect=[2, 100],
    )
    def test_given_fgs_plan_finished_when_zocalo_results_event_then_expected_comment_deposited(
        self, mock_time, dummy_rotation_data_collection_group_info
    ):
        ispyb_handler = GridscanISPyBCallback(
            param_type=GridCommonWithHyperionDetectorParams,
        )

        ispyb_handler.activity_gated_start(td.test_gridscan3d_start_document)  # type:ignore

        ispyb_handler.activity_gated_start(td.test_do_fgs_start_document)  # type:ignore
        ispyb_handler.activity_gated_stop(td.test_do_fgs_gridscan_stop_document)

        ispyb_handler.activity_gated_descriptor(
            td.test_descriptor_document_zocalo_reading
        )
        ispyb_handler.data_collection_group_info = (
            dummy_rotation_data_collection_group_info
        )
        ispyb_handler.activity_gated_event(td.test_zocalo_reading_event)

        assert (
            ispyb_handler.ispyb.append_to_comment.call_args.args[1]  # type:ignore
            == "Zocalo processing took 98.00 s. Zocalo found no crystals in this gridscan."
        )



from enum import Enum

import bluesky.plan_stubs as bps
import bluesky.preprocessors as bpp
import numpy as np
import pydantic
from blueapi.core import BlueskyContext
from dodal.devices.attenuator.attenuator import BinaryFilterAttenuator
from dodal.devices.xspress3.xspress3 import Xspress3
from dodal.devices.zebra.zebra_controlled_shutter import ZebraShutter, ZebraShutterState

from mx_bluesky.common.utils.context import device_composite_from_context
from mx_bluesky.common.utils.log import LOGGER


class AttenuationOptimisationFailedException(Exception):
    pass


class Direction(Enum):
    POSITIVE = "positive"
    NEGATIVE = "negative"


@pydantic.dataclasses.dataclass(config={"arbitrary_types_allowed": True})
class OptimizeAttenuationComposite:
    """All devices which are directly or indirectly required by this plan"""

    attenuator: BinaryFilterAttenuator
    sample_shutter: ZebraShutter
    xspress3mini: Xspress3


def create_devices(context: BlueskyContext) -> OptimizeAttenuationComposite:
    return device_composite_from_context(context, OptimizeAttenuationComposite)


def check_parameters(
    target,
    upper_count_limit,
    lower_count_limit,
    default_high_roi,
    default_low_roi,
    initial_transmission,
    upper_transmission,
    lower_transmission,
):
    if target < lower_count_limit or target > upper_count_limit:
        raise (
            ValueError(
                f"Target {target} is outside of lower and upper bounds: {lower_count_limit} to {upper_count_limit}"
            )
        )

    if default_high_roi < default_low_roi:
        raise ValueError(
            f"Upper roi {default_high_roi} must be greater than lower roi {default_low_roi}"
        )

    if upper_transmission < lower_transmission:
        raise ValueError(
            f"Upper transmission limit {upper_transmission} must be greater than lower tranmission limit {lower_transmission}"
        )

    if not upper_transmission >= initial_transmission >= lower_transmission:
        raise ValueError(
            f"initial transmission {initial_transmission} is outside range {lower_transmission} - {upper_transmission}"
        )


def is_counts_within_target(total_count, lower_count_limit, upper_count_limit) -> bool:
    if lower_count_limit <= total_count and total_count <= upper_count_limit:
        return True
    else:
        return False


def calculate_new_direction(direction: Direction, deadtime, deadtime_threshold):
    if direction == Direction.POSITIVE:
        if deadtime > deadtime_threshold:
            direction = Direction.NEGATIVE
            LOGGER.info(
                "Found tranmission to go above deadtime threshold. Reducing transmission..."
            )
    return direction


def deadtime_calc_new_transmission(
    direction: Direction,
    transmission: float,
    increment: float,
    upper_transmission_limit: float,
    lower_transmission_limit: float,
) -> float:
    """Calculate the new transmission value based on the current direction and increment. Raise error if transmission is too low.

    Args:
        direction (Direction):
        If positive, increase transmission by a factor of the increment. If negative, divide it

        transmission (float):
        Current transmission value

        increment (float):
        Factor to multiply or divide transmission by

        upper_transmission_limit (float):
        Maximum allowed transmission, in order to protect sample.

        lower_transmission_limit (float):
        Minimum expected transmission. Raise an error if transmission goes lower.

    Raises:
        AttenuationOptimisationFailedException:
        This error is thrown if the transmission goes below the expected value or if the maximum cycles are reached

    Returns:
        transmission (float): Optimised transmission value
    """
    if direction == Direction.POSITIVE:
        transmission *= increment
        if transmission > upper_transmission_limit:
            transmission = upper_transmission_limit
    else:
        transmission /= increment
    if transmission < lower_transmission_limit:
        raise AttenuationOptimisationFailedException(
            "Calculated transmission is below expected limit"
        )
    return transmission


def do_device_optimise_iteration(
    composite: OptimizeAttenuationComposite,
    transmission,
):
    def close_shutter():
        yield from bps.abs_set(
            composite.sample_shutter, ZebraShutterState.CLOSE, wait=True
        )

    @bpp.finalize_decorator(close_shutter)
    def open_and_run():
        """Set transmission, set number of images on xspress3mini, arm xspress3mini"""
        yield from bps.abs_set(
            composite.attenuator, transmission, group="set_transmission"
        )
        yield from bps.abs_set(composite.xspress3mini.set_num_images, 1, wait=True)
        yield from bps.abs_set(
            composite.sample_shutter, ZebraShutterState.OPEN, wait=True
        )
        yield from bps.stage(composite.xspress3mini, wait=True)
        yield from bps.unstage(composite.xspress3mini, wait=True)

    yield from open_and_run()


def is_deadtime_optimised(
    deadtime: float,
    deadtime_threshold: float,
    transmission: float,
    upper_transmission_limit: float,
    direction: Direction,
) -> bool:
    if direction == Direction.POSITIVE:
        if transmission == upper_transmission_limit:
            LOGGER.warning(
                f"Deadtime {deadtime} is above threshold {deadtime_threshold} at maximum transmission {upper_transmission_limit}. Using maximum transmission\
                        as optimised value."
            )
            return True
    # Once direction is flipped and deadtime goes back above threshold, we consider attenuation to be optimised.
    else:
        if deadtime <= deadtime_threshold:
            return True
    return False


def deadtime_optimisation(
    composite: OptimizeAttenuationComposite,
    transmission: float,
    increment: float,
    deadtime_threshold: float,
    max_cycles: int,
    upper_transmission_limit: float,
    lower_transmission_limit: float,
):
    """Optimises the attenuation for the Xspress3Mini based on the detector deadtime

    Deadtime is the time after each event during which the detector cannot record another event. This loop adjusts the transmission of the attenuator
    and checks the deadtime until the percentage deadtime is below the accepted threshold. To protect the sample, the transmission has a maximum value

    Here we use the percentage deadtime - the percentage of time to which the detector is unable to process events.

    This algorithm gradually increases the transmission until the percentage deadtime goes beneath the specified threshold. It then increases
    the transmission and stops when the deadtime goes above the threshold. A smaller increment will provide a better optimised value, but take more
    cycles to complete.

    Args:
        composite: (OptimizeAttenuationComposite) Devices required to optimise attenuation

        transmission: (float)
        The initial transmission value to use for the optimising

        increment: (float)
        The factor to increase / decrease the transmission by each iteration

        deadtime_threshold: (float)
        The maximum acceptable percentage deadtime

        max_cycles: (int)
        The maximum number of iterations before an error is thrown

        upper_transmission_limit (float):
        Maximum allowed transmission, in order to protect sample.

        lower_transmission_limit (float):
        Minimum expected transmission. Raise an error if transmission goes lower.

    Raises:
        AttenuationOptimisationFailedException:
        This error is thrown if the transmission goes below the expected value or the maximum cycles are reached

    Returns:
        optimised_transmission: (float)
        The final transmission value which produces an acceptable deadtime
    """

    direction = Direction.POSITIVE
    LOGGER.info(f"Target deadtime is {deadtime_threshold}")
    optimised_transmission: float = 0
    for cycle in range(0, max_cycles):
        yield from do_device_optimise_iteration(composite, transmission)

        total_time = yield from bps.rd(composite.xspress3mini.channels[1].total_time)
        reset_ticks = yield from bps.rd(composite.xspress3mini.channels[1].reset_ticks)

        LOGGER.info(f"Current total time = {total_time}")
        LOGGER.info(f"Current reset ticks = {reset_ticks}")
        deadtime = 0

        """
            The reset ticks PV stops ticking while the detector is unable to process events, so the absolute difference between the total time and the
            reset ticks time gives the deadtime in unit time. Divide by total time to get it as a percentage.
        """

        if total_time != reset_ticks:
            deadtime = 1 - abs(total_time - reset_ticks) / (total_time)

        LOGGER.info(f"Deadtime is now at {deadtime}")

        # Check if new deadtime is OK

        if is_deadtime_optimised(
            deadtime,
            deadtime_threshold,
            transmission,
            upper_transmission_limit,
            direction,
        ):
            optimised_transmission = transmission
            break

        if cycle == max_cycles - 1:
            raise AttenuationOptimisationFailedException(
                f"Unable to optimise attenuation after maximum cycles.\
                                                            Deadtime did not get lower than threshold: {deadtime_threshold} in maximum cycles {max_cycles}"
            )

        direction = calculate_new_direction(direction, deadtime, deadtime_threshold)

        transmission = deadtime_calc_new_transmission(
            direction,
            transmission,
            increment,
            upper_transmission_limit,
            lower_transmission_limit,
        )

    return optimised_transmission


def total_counts_optimisation(
    composite: OptimizeAttenuationComposite,
    transmission: float,
    low_roi: int,
    high_roi: int,
    lower_count_limit: float,
    upper_count_limit: float,
    target_count: float,
    max_cycles: int,
    upper_transmission_limit: float,
    lower_transmission_limit: float,
):
    """Optimises the attenuation for the Xspress3Mini based on the total counts

    This loop adjusts the transmission of the attenuator and checks the total counts of the detector until the total counts as in the acceptable range,
    defined by the lower and upper limit. To protect the sample, the transmission has a maximum value of 10%.

    Args:
        composite: (OptimizeAttenuationComposite) Devices required to optimise attenuation

        transmission: (float)
        The initial transmission value to use for the optimising

        low_roi: (float)
        Lower region of interest at which to include in the counts

        high_roi: (float)
        Upper region of interest at which to include in the counts

        lower_count_limit: (float)
        The lowest acceptable value for count

        upper_count_limit: (float)
        The highest acceptable value for count

        target_count: (int)
        The ideal number of target counts - used to calculate the transmission for the subsequent iteration.

        max_cycles: (int)
        The maximum number of iterations before an error is thrown

        upper_transmission_limit: (float)
        The maximum allowed value for the transmission

        lower_transmission_limit: (float)
        The minimum allowed value for the transmission

    Returns:
        optimised_transmission: (float)
        The final transmission value which produces an acceptable total_count value
    """

    LOGGER.info("Using total count optimisation")
    optimised_transmission: float = 0
    for cycle in range(0, max_cycles):
        LOGGER.info(
            f"Setting transmission to {transmission} for attenuation optimisation cycle {cycle}"
        )

        yield from do_device_optimise_iteration(composite, transmission)

        data = np.array(
            (yield from bps.rd(composite.xspress3mini.dt_corrected_latest_mca[1]))
        )
        total_count = sum(data[int(low_roi) : int(high_roi)])
        LOGGER.info(f"Total count is {total_count}")

        if is_counts_within_target(total_count, lower_count_limit, upper_count_limit):
            optimised_transmission = transmission
            LOGGER.info(
                f"Total count is within accepted limits: {lower_count_limit}, {total_count}, {upper_count_limit}"
            )
            break
        elif transmission == upper_transmission_limit:
            LOGGER.warning(
                f"Total count is not within limits: {lower_count_limit} <= {total_count} <= {upper_count_limit}\
                    after using maximum transmission {upper_transmission_limit}. Continuing\
                    with maximum transmission as optimised value..."
            )
            optimised_transmission = transmission
            break

        else:
            transmission = (target_count / (total_count)) * transmission
            if transmission > upper_transmission_limit:
                transmission = upper_transmission_limit
            elif transmission < lower_transmission_limit:
                raise AttenuationOptimisationFailedException(
                    f"Transmission has gone below lower threshold {lower_transmission_limit}"
                )

        if cycle == max_cycles - 1:
            raise AttenuationOptimisationFailedException(
                f"Unable to optimise attenuation after maximum cycles.\
                                                            Total count is not within limits: {lower_count_limit} <= {total_count}\
                                                                <= {upper_count_limit}"
            )

    return optimised_transmission


def optimise_attenuation_plan(
    composite: OptimizeAttenuationComposite,
    collection_time=1,  # Comes from self.parameters.acquisitionTime in fluorescence_spectrum.py
    optimisation_type="deadtime",
    low_roi=100,
    high_roi=2048,
    upper_transmission_limit=0.1,
    lower_transmission_limit=1.0e-6,
    initial_transmission=0.1,
    target_count=20000,
    lower_count_limit=20000,
    upper_count_limit=50000,
    max_cycles=10,
    increment=2,
    deadtime_threshold=0.002,
):
    check_parameters(
        target_count,
        upper_count_limit,
        lower_count_limit,
        high_roi,
        low_roi,
        initial_transmission,
        upper_transmission_limit,
        lower_transmission_limit,
    )

    yield from bps.abs_set(
        composite.xspress3mini.acquire_time, collection_time, wait=True
    )  # Don't necessarily need to wait here
    optimised_transmission: float = 0
    # Do the attenuation optimisation using count threshold
    if optimisation_type == "total_counts":
        LOGGER.info(
            f"Starting Xspress3Mini total counts optimisation routine \nOptimisation will be performed across ROI channels {low_roi} - {high_roi}"
        )

        optimised_transmission = yield from total_counts_optimisation(
            composite,
            initial_transmission,
            low_roi,
            high_roi,
            lower_count_limit,
            upper_count_limit,
            target_count,
            max_cycles,
            upper_transmission_limit,
            lower_transmission_limit,
        )

    elif optimisation_type == "deadtime":
        LOGGER.info(
            f"Starting Xspress3Mini deadtime optimisation routine \nOptimisation will be performed across ROI channels {low_roi} - {high_roi}"
        )
        optimised_transmission = yield from deadtime_optimisation(
            composite,
            initial_transmission,
            upper_transmission_limit,
            lower_transmission_limit,
            increment,
            deadtime_threshold,
            max_cycles,
        )

    yield from bps.abs_set(
        composite.attenuator,
        optimised_transmission,
        group="set_transmission",
        wait=True,
    )

    return optimised_transmission



import builtins
import dataclasses
import time
from abc import ABC
from typing import Literal

from bluesky.protocols import Readable, Reading
from event_model import DataKey


@dataclasses.dataclass(frozen=True)
class AbstractEvent(Readable, ABC):
    """An abstract superclass that can be extended to provide lightweight software events
    for bluesky plans, without having to incur the overhead of creating ophyd-async devices
    specifically for the purpose.

    The currently supported types for field annotations in the event are ``str``, ``int``, ``float``, ``bool``

    In future array types may be supported.

    Examples:
        Subclasses should extend this class and decorate with::

            @dataclasses.dataclass(frozen=True)

        To raise an event, simply construct the event and then ``read`` it as you would a device::

            yield from bps.create("MY_EVENT_NAME")
            my_event = MyEvent(an_int=1)
            yield from bps.read(my_event)
            yield from bps.save()
    """

    def read(self) -> dict[str, Reading]:
        return {
            f.name: AbstractEvent._reading_from_value(getattr(self, f.name))
            for f in dataclasses.fields(self)
        }

    def describe(self) -> dict[str, DataKey]:
        return {
            f.name: DataKey(dtype=AbstractEvent._dtype_of(f.type), shape=[], source="")
            for f in dataclasses.fields(self)
        }

    @classmethod
    def _reading_from_value(cls, value):
        return Reading(timestamp=time.time(), value=value)

    @classmethod
    def _dtype_of(cls, t) -> Literal["string", "number", "boolean", "integer"]:
        match t:
            case builtins.str:
                return "string"
            case builtins.bool:
                return "boolean"
            case builtins.int:
                return "integer"
            case builtins.float:
                return "number"
        # TODO array support
        raise ValueError(f"Unsupported type for AbstractEvent: {t}")

    @property
    def name(self) -> str:
        return type(self).__name__



import argparse
import subprocess
from os import environ
from pathlib import Path

from mx_bluesky.beamlines.i24.serial.log import SSX_LOGGER
from mx_bluesky.beamlines.i24.serial.parameters import SSXType


def _parse_input(expt: SSXType):
    parser = argparse.ArgumentParser(description=f"Run a {expt} collection.")
    parser.add_argument("-t", "--test", action="store_true", help="Run in test mode.")
    args = parser.parse_args()
    return args


def get_location(default: str = "dev") -> str:
    return environ.get("BEAMLINE") or default


def get_edm_path() -> Path:
    return Path(__file__).parents[5] / "edm_serial"


def _get_file_path() -> Path:
    return Path(__file__).parent


def run_extruder():
    args = _parse_input(SSXType.EXTRUDER)
    loc = get_location()
    SSX_LOGGER.info(f"Running on {loc}.")
    edm_path = get_edm_path()
    filepath = _get_file_path()
    test_mode = "--test" if args.test else ""
    SSX_LOGGER.debug(f"Running {filepath}/run_extruder.sh")
    subprocess.run(
        ["bash", filepath / "run_extruder.sh", edm_path.as_posix(), test_mode]
    )


def run_fixed_target():
    args = _parse_input(SSXType.FIXED)
    loc = get_location()
    SSX_LOGGER.info(f"Running on {loc}.")
    edm_path = get_edm_path()
    filepath = _get_file_path()
    test_mode = "--test" if args.test else ""
    SSX_LOGGER.debug(f"Running {filepath}/run_fixed_target.sh")
    subprocess.run(
        ["bash", filepath / "run_fixed_target.sh", edm_path.as_posix(), test_mode]
    )
